# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from sahara import context
from sahara.db.templates import api as template_api
from sahara.db.templates import utils as u
from sahara.tests.unit.conductor import base
from sahara.tests.unit.db.templates import common as c


class Config(c.Config):
    def __init__(self, option_values=None):
        option_values = option_values or {}
        if "name" not in option_values:
            option_values["name"] = "delete"
        super(Config, self).__init__(option_values)


class TemplateDeleteTestCase(base.ConductorManagerTestCase):
    def setUp(self):
        super(TemplateDeleteTestCase, self).setUp()
        self.logger = c.Logger()
        template_api.set_logger(self.logger)

    def test_node_group_template_delete_by_id(self):
        self.logger.clear_log()
        self.setup_context(tenant_id=None)
        ctx = context.ctx()
        t = self.api.node_group_template_create(ctx, c.SAMPLE_NGT)

        option_values = {"tenant_id": t["tenant_id"],
                         "id": t["id"]}
        template_api.set_conf(Config(option_values))

        template_api.do_node_group_template_delete_by_id()
        msg = 'Deleted node group template {info}'.format(
            info=u.name_and_id(t))
        self.assertIn(msg, self.logger.infos)

        t = self.api.node_group_template_get(ctx, t["id"])
        self.assertIsNone(t)

    def test_node_group_template_delete_by_id_skipped(self):
        self.logger.clear_log()
        self.setup_context(tenant_id=None)
        ctx = context.ctx()
        template_values = copy.copy(c.SAMPLE_NGT)
        template_values["is_default"] = False
        t = self.api.node_group_template_create(ctx, template_values)

        option_values = {"tenant_id": t["tenant_id"],
                         "id": t["id"]}
        template_api.set_conf(Config(option_values))

        template_api.do_node_group_template_delete_by_id()
        msg = ("Deletion of node group template {info} skipped, "
               "not a default template".format(info=u.name_and_id(t)))
        self.assertIn(msg, self.logger.warnings)

        t = self.api.node_group_template_get(ctx, t["id"])
        self.assertIsNotNone(t)

    def test_node_group_template_delete_bad_id(self):
        self.logger.clear_log()
        option_values = {"tenant_id": 1,
                         "id": "badid"}
        template_api.set_conf(Config(option_values))
        template_api.do_node_group_template_delete_by_id()
        msg = ("Deletion of node group template {id} failed, "
               "no such template".format(id=option_values["id"]))
        self.assertIn(msg, self.logger.warnings)

    def test_node_group_template_delete_by_name(self):
        self.logger.clear_log()
        ctx = context.ctx()
        t = self.api.node_group_template_create(ctx, c.SAMPLE_NGT)

        option_values = {"tenant_id": t["tenant_id"],
                         "template_name": t["name"]}
        template_api.set_conf(Config(option_values))

        template_api.do_node_group_template_delete()
        msg = 'Deleted node group template {info}'.format(
            info=u.name_and_id(t))
        self.assertIn(msg, self.logger.infos)

        t = self.api.node_group_template_get(ctx, t["id"])
        self.assertIsNone(t)

    def test_node_group_template_delete_by_name_skipped(self):
        self.logger.clear_log()
        ctx = context.ctx()
        template_values = copy.copy(c.SAMPLE_NGT)
        template_values["is_default"] = False
        t = self.api.node_group_template_create(ctx, template_values)

        option_values = {"tenant_id": t["tenant_id"],
                         "template_name": t["name"]}
        template_api.set_conf(Config(option_values))

        template_api.do_node_group_template_delete()
        msg = ("Deletion of node group template {name} failed, "
               "no such template".format(name=t["name"]))
        self.assertIn(msg, self.logger.warnings)

        t = self.api.node_group_template_get(ctx, t["id"])
        self.assertIsNotNone(t)

    def test_node_group_template_delete_in_use(self):
        self.logger.clear_log()
        ctx = context.ctx()
        t = self.api.node_group_template_create(ctx, c.SAMPLE_NGT)

        # Make a cluster that references the node group template
        cluster_values = copy.deepcopy(c.SAMPLE_CLUSTER)
        cluster_values["node_groups"][0]["node_group_template_id"] = t["id"]
        cl = self.api.cluster_create(ctx, cluster_values)

        # Make a cluster template that references the node group template
        cluster_temp_values = copy.deepcopy(c.SAMPLE_CLT)
        cluster_temp_values["node_groups"] = cluster_values["node_groups"]
        clt = self.api.cluster_template_create(ctx, cluster_temp_values)

        # Set up the expected messages
        msgs = ["Node group template {info} in use "
                "by clusters {clusters}".format(
                    info=u.name_and_id(t), clusters=[cl["name"]])]

        msgs += ["Node group template {info} in use "
                 "by cluster templates {cluster_temps}".format(
                     info=u.name_and_id(t), cluster_temps=[clt["name"]])]

        msgs += ["Deletion of node group template {info} failed".format(
            info=u.name_and_id(t))]

        # Check delete by name
        option_values = {"tenant_id": t["tenant_id"],
                         "template_name": t["name"]}
        template_api.set_conf(Config(option_values))
        template_api.do_node_group_template_delete()
        for msg in msgs:
            self.assertIn(msg, self.logger.warnings)
        self.logger.clear_log()

        # Check again with delete by id
        option_values = {"tenant_id": t["tenant_id"],
                         "id": t["id"]}
        template_api.set_conf(Config(option_values))
        template_api.do_node_group_template_delete_by_id()
        for msg in msgs:
            self.assertIn(msg, self.logger.warnings)
        self.logger.clear_log()

    def test_cluster_template_delete_by_id(self):
        self.logger.clear_log()
        self.setup_context(tenant_id=None)
        ctx = context.ctx()
        t = self.api.cluster_template_create(ctx, c.SAMPLE_CLT)

        option_values = {"tenant_id": t["tenant_id"],
                         "id": t["id"]}
        template_api.set_conf(Config(option_values))

        template_api.do_cluster_template_delete_by_id()
        msg = 'Deleted cluster template {info}'.format(
            info=u.name_and_id(t))
        self.assertIn(msg, self.logger.infos)

        t = self.api.cluster_template_get(ctx, t["id"])
        self.assertIsNone(t)

    def test_cluster_template_delete_by_id_skipped(self):
        self.logger.clear_log()
        ctx = context.ctx()
        template_values = copy.copy(c.SAMPLE_CLT)
        template_values["is_default"] = False
        t = self.api.cluster_template_create(ctx, template_values)

        option_values = {"tenant_id": t["tenant_id"],
                         "id": t["id"]}
        template_api.set_conf(Config(option_values))
        template_api.do_cluster_template_delete_by_id()
        msg = ("Deletion of cluster template {info} skipped, "
               "not a default template".format(info=u.name_and_id(t)))
        self.assertIn(msg, self.logger.warnings)

        t = self.api.cluster_template_get(ctx, t["id"])
        self.assertIsNotNone(t)

    def test_cluster_template_delete_bad_id(self):
        self.logger.clear_log()
        option_values = {"tenant_id": 1,
                         "id": "badid"}
        template_api.set_conf(Config(option_values))
        template_api.do_cluster_template_delete_by_id()
        msg = ("Deletion of cluster template {id} failed, "
               "no such template".format(id=option_values["id"]))
        self.assertIn(msg, self.logger.warnings)

    def test_cluster_template_delete_by_name(self):
        self.logger.clear_log()
        ctx = context.ctx()
        t = self.api.cluster_template_create(ctx, c.SAMPLE_NGT)

        option_values = {"tenant_id": t["tenant_id"],
                         "template_name": t["name"]}
        template_api.set_conf(Config(option_values))

        template_api.do_cluster_template_delete()
        msg = 'Deleted cluster template {info}'.format(
            info=u.name_and_id(t))
        self.assertIn(msg, self.logger.infos)

        t = self.api.cluster_template_get(ctx, t["id"])
        self.assertIsNone(t)

    def test_cluster_template_delete_by_name_skipped(self):
        self.logger.clear_log()
        ctx = context.ctx()
        template_values = copy.copy(c.SAMPLE_NGT)
        template_values["is_default"] = False
        t = self.api.cluster_template_create(ctx, template_values)

        option_values = {"tenant_id": t["tenant_id"],
                         "template_name": t["name"]}
        template_api.set_conf(Config(option_values))

        template_api.do_cluster_template_delete()
        msg = ("Deletion of cluster template {name} failed, "
               "no such template".format(name=t["name"]))
        self.assertIn(msg, self.logger.warnings)

        t = self.api.cluster_template_get(ctx, t["id"])
        self.assertIsNotNone(t)

    def test_cluster_template_delete_in_use(self):
        self.logger.clear_log()
        ctx = context.ctx()
        t = self.api.cluster_template_create(ctx, c.SAMPLE_CLT)

        # Make a cluster that references the cluster template
        cluster_values = copy.deepcopy(c.SAMPLE_CLUSTER)
        cluster_values["cluster_template_id"] = t["id"]
        del cluster_values["node_groups"]
        cl = self.api.cluster_create(ctx, cluster_values)

        # Set up the expected messages
        msgs = ["Cluster template {info} in use "
                "by clusters {clusters}".format(
                    info=u.name_and_id(t), clusters=[cl["name"]])]

        msgs += ["Deletion of cluster template {info} failed".format(
            info=u.name_and_id(t))]

        # Check delete by name
        option_values = {"tenant_id": t["tenant_id"],
                         "template_name": t["name"]}
        template_api.set_conf(Config(option_values))
        template_api.do_cluster_template_delete()
        for msg in msgs:
            self.assertIn(msg, self.logger.warnings)
        self.logger.clear_log()

        # Check again with delete by id
        option_values = {"tenant_id": t["tenant_id"],
                         "id": t["id"]}
        template_api.set_conf(Config(option_values))
        template_api.do_cluster_template_delete_by_id()
        for msg in msgs:
            self.assertIn(msg, self.logger.warnings)
        self.logger.clear_log()

    def _make_templates(self, ctx, name, plugin_name, plugin_version):
        # Make a node group template
        values = copy.copy(c.SAMPLE_NGT)
        values["name"] = "ngt_" + name
        values["plugin_name"] = plugin_name
        values["hadoop_version"] = plugin_version
        ngt = self.api.node_group_template_create(ctx, values)

        # Make a cluster template that references the node group template
        values = copy.deepcopy(c.SAMPLE_CLT)
        values["name"] = "clt_" + name
        values["plugin_name"] = plugin_name
        values["hadoop_version"] = plugin_version
        values["node_groups"][0]["node_group_template_id"] = ngt["id"]
        clt = self.api.cluster_template_create(ctx, values)

        return ngt, clt

    def test_do_delete(self):
        self.logger.clear_log()
        ctx = context.ctx()

        # Make some plugins to delete
        ngt, clt = self._make_templates(ctx, "first", "plugin", "v1")

        # Make some more for the same plugin, different version
        ngt2, clt2 = self._make_templates(ctx, "second", "plugin", "v2")

        # Make another set for a different plugin, overlapping version
        safe_ngt, safe_clt = self._make_templates(ctx, "third", "plugin2",
                                                  "v1")

        # Run a delete by plugin name/version for the first set
        option_values = {"tenant_id": ngt["tenant_id"],
                         "plugin_name": [ngt["plugin_name"]],
                         "plugin_version": [ngt["hadoop_version"]]}
        template_api.set_conf(Config(option_values))

        # Should delete clt and then ngt, check for messages in order
        template_api.do_delete()
        msgs = ["Deleted cluster template {info}".format(
            info=u.name_and_id(clt))]
        msgs += ["Deleted node group template {info}".format(
            info=u.name_and_id(ngt))]

        self.assertEqual(msgs, self.logger.infos)
        self.assertIsNone(self.api.node_group_template_get(ctx, ngt["id"]))
        self.assertIsNone(self.api.cluster_template_get(ctx, clt["id"]))

        # Make sure the other templates are still there
        self.assertIsNotNone(self.api.node_group_template_get(ctx,
                                                              ngt2["id"]))
        self.assertIsNotNone(self.api.cluster_template_get(ctx,
                                                           clt2["id"]))
        self.assertIsNotNone(self.api.node_group_template_get(ctx,
                                                              safe_ngt["id"]))
        self.assertIsNotNone(self.api.cluster_template_get(ctx,
                                                           safe_clt["id"]))

        # Run delete again for the plugin but with no version specified
        self.logger.clear_log()
        option_values = {"tenant_id": ngt2["tenant_id"],
                         "plugin_name": [ngt2["plugin_name"]],
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))

        # Should delete clt2 and then ngt2, check for messages in order
        template_api.do_delete()
        msgs = ["Deleted cluster template {info}".format(
            info=u.name_and_id(clt2))]
        msgs += ["Deleted node group template {info}".format(
            info=u.name_and_id(ngt2))]

        self.assertEqual(msgs, self.logger.infos)
        self.assertIsNone(self.api.node_group_template_get(ctx, ngt2["id"]))
        self.assertIsNone(self.api.cluster_template_get(ctx, clt2["id"]))

        # Make sure the other templates are still there
        self.assertIsNotNone(self.api.node_group_template_get(ctx,
                                                              safe_ngt["id"]))
        self.assertIsNotNone(self.api.cluster_template_get(ctx,
                                                           safe_clt["id"]))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import mock

from oslo_utils import uuidutils

from sahara.service.api import v10 as api
from sahara.service.validations import cluster_template_schema as ct_schema
from sahara.service.validations import cluster_templates as ct
from sahara.tests.unit.service.validation import utils as u


class TestClusterTemplateCreateValidation(u.ValidationTestCase):
    def setUp(self):
        super(TestClusterTemplateCreateValidation, self).setUp()
        self._create_object_fun = ct.check_cluster_template_create
        self.scheme = ct_schema.CLUSTER_TEMPLATE_SCHEMA
        api.plugin_base.setup_plugins()

    def test_cluster_template_create_v_cluster_configs(self):
        self._assert_cluster_configs_validation()

    def test_cluster_template_create_v_ng(self):
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "fake",
                'hadoop_version': "0.1",
                'node_groups': [
                    {'name': 'a'}
                ]
            },
            bad_req_i=(1, 'VALIDATION_ERROR',
                       "node_groups[0]: {'name': 'a'} is not valid under "
                       "any of the given schemas")
        )
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "fake",
                'hadoop_version': "0.1",
                'node_groups': [
                    {'name': 'a',
                     'flavor_id': '42'}
                ]
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       "node_groups[0]: {'name': 'a', 'flavor_id': '42'} "
                       "is not valid under any of the given schemas")

        )
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "fake",
                'hadoop_version': "0.1",
                'node_groups': [
                    {'name': 'a',
                     'flavor_id': '42',
                     'node_processes': ['namenode']}
                ]
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       "node_groups[0]: {'node_processes': ['namenode'], "
                       "'name': 'a', "
                       "'flavor_id': '42'} "
                       "is not valid under any of the given schemas")
        )
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "fake",
                'hadoop_version': "0.1",
                'node_groups': [
                    {
                        'name': 'a',
                        'flavor_id': '42',
                        'node_processes': ['namenode'],
                        'count': 1
                    },
                    {
                        "node_group_template_id": "550e8400-e29b-41d4-a716-"
                                                  "446655440000",
                        "name": "a",
                        'count': 2
                    }
                ]
            },
            bad_req_i=(1, "INVALID_DATA",
                       "Duplicates in node group names are detected")
        )

    def test_cluster_template_create_v_ng_templates(self):
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "fake",
                'hadoop_version': "0.1",
                'node_groups': [
                    {
                        "node_group_template_id": "",
                        "name": "test",
                    }
                ]
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       "node_groups[0]: {'node_group_template_id': '', "
                       "'name': 'test'} "
                       "is not valid under any of the given schemas")

        )
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "fake",
                'hadoop_version': "0.1",
                'node_groups': [
                    {
                        "node_group_template_id": "test",
                        "name": "test",
                        'count': 3
                    }
                ]
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       "node_groups[0]: {'count': 3, "
                       "'node_group_template_id': 'test', "
                       "'name': 'test'} "
                       "is not valid under any of the given schemas")

        )

    def test_cluster_template_create_v_ng_templates_right(self):
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "fake",
                'hadoop_version': "0.1",
                'node_groups': [
                    {
                        "node_group_template_id": "550e8400-e29b-41d4-a716-"
                                                  "446655440000",
                        "name": "test",
                        'count': 3
                    }
                ],
                'domain_name': 'domain.org.'
            },

        )

    def test_cluster_template_create_v_name_base(self):
        data = {
            'name': "testname",
            'plugin_name': "fake",
            'hadoop_version': "0.1"
        }
        self._assert_valid_name_hostname_validation(data)

    def test_cluster_template_create_v_types(self):
        data = {
            'name': "testname",
            'plugin_name': "fake",
            'hadoop_version': "0.1"
        }
        self._assert_types(data)

    def test_cluster_template_create_v_required(self):
        self._assert_create_object_validation(
            data={},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'name' is a required property")
        )
        self._assert_create_object_validation(
            data={
                'name': 'test-name'
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'plugin_name' is a required property")
        )
        self._assert_create_object_validation(
            data={
                'name': 'testname',
                'plugin_name': 'fake'
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'hadoop_version' is a required property")
        )

    @mock.patch("sahara.service.validations.base.check_all_configurations")
    @mock.patch("sahara.service.validations.base.check_network_exists")
    @mock.patch("sahara.service.validations.base.check_required_image_tags")
    @mock.patch("sahara.service.validations.base.check_image_registered")
    def test_cluster_template_create_v_right(self, image_reg, image_tags,
                                             net_exists, all_configs):
        self._assert_create_object_validation(
            data={
                'name': 'testname',
                'plugin_name': 'fake',
                'hadoop_version': '0.1',
                'default_image_id': uuidutils.generate_uuid(),
                'cluster_configs': {
                    "service_1": {
                        "config_1": "value_1"
                    }
                },
                'node_groups': [
                    {
                        "node_group_template_id": "550e8400-e29b-41d4-a716-"
                                                  "446655440000",
                        "name": "charlie",
                        'count': 3
                    }
                ],
                'anti_affinity': ['datanode'],
                'description': 'my template',
                'neutron_management_network': uuidutils.generate_uuid(),
                'domain_name': 'domain.org.'
            })

    @mock.patch("sahara.service.validations.base.check_network_exists")
    @mock.patch("sahara.service.validations.base.check_required_image_tags")
    @mock.patch("sahara.service.validations.base.check_image_registered")
    def test_cluster_template_create_v_right_nulls(self, image_reg, image_tags,
                                                   net_exists):
        self._assert_create_object_validation(
            data={
                'name': 'testname',
                'plugin_name': 'fake',
                'hadoop_version': '0.1',
                'default_image_id': None,
                'cluster_configs': None,
                'node_groups': None,
                'anti_affinity': None,
                'description': None,
                'neutron_management_network': None,
                'domain_name': None
            })

    def test_cluster_template_create_v_plugin_name_exists(self):
        self._assert_create_object_validation(
            data={
                'name': "test-name",
                'plugin_name': "wrong_plugin",
                'hadoop_version': "0.1",
            },
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Sahara doesn't contain plugin "
                       "with name 'wrong_plugin'")
        )

    def test_cluster_template_create_v_unique_cl(self):
        data = {
            'name': 'test',
            'plugin_name': 'fake',
            'hadoop_version': '0.1'
        }
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, 'NAME_ALREADY_EXISTS',
                       "Cluster template with name 'test' already exists")
        )

    def test_cluster_template_wrong_neutron_mngmt_net(self):
        data = {
            'name': 'test-template',
            'plugin_name': 'fake',
            'hadoop_version': '0.1',
            'neutron_management_network': '53a36917-ab9f-4589'
                                          '-94ce-b6df85a68332'
        }
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, 'NOT_FOUND', "Network 53a36917-ab9f-4589-"
                                       "94ce-b6df85a68332 not found")
        )

    def test_cluster_create_v_default_image_required_tags(self):
        self._assert_cluster_default_image_tags_validation()



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging

from sahara import context
from sahara.utils import rpc as messaging

LOG = logging.getLogger(__name__)

SERVICE = 'sahara'
CLUSTER_EVENT_TEMPLATE = "sahara.cluster.%s"
HEALTH_EVENT_TYPE = CLUSTER_EVENT_TEMPLATE % "health"

notifier_opts = [
    cfg.StrOpt('level',
               default='INFO',
               deprecated_name='notification_level',
               deprecated_group='DEFAULT',
               help='Notification level for outgoing notifications'),
    cfg.StrOpt('publisher_id',
               deprecated_name='notification_publisher_id',
               deprecated_group='DEFAULT',
               help='Identifier of the publisher')
]

notifier_opts_group = 'oslo_messaging_notifications'

CONF = cfg.CONF
CONF.register_opts(notifier_opts, group=notifier_opts_group)


def _get_publisher():
    publisher_id = CONF.oslo_messaging_notifications.publisher_id
    if publisher_id is None:
        publisher_id = SERVICE
    return publisher_id


def _notify(event_type, body):
    LOG.debug("Notification about cluster is going to be sent. Notification "
              "type={type}".format(type=event_type))
    ctx = context.ctx()
    level = CONF.oslo_messaging_notifications.level

    body.update({'project_id': ctx.tenant_id, 'user_id': ctx.user_id})
    client = messaging.get_notifier(_get_publisher())

    method = getattr(client, level.lower())
    method(ctx, event_type, body)


def _health_notification_body(cluster, health_check):
    verification = cluster.verification
    return {
        'cluster_id': cluster.id,
        'cluster_name': cluster.name,
        'verification_id': verification['id'],
        'health_check_status': health_check['status'],
        'health_check_name': health_check['name'],
        'health_check_description': health_check['description'],
        'created_at': health_check['created_at'],
        'updated_at': health_check['updated_at']
    }


def status_notify(cluster_id, cluster_name, cluster_status, ev_type):
    """Sends notification about creating/updating/deleting cluster."""
    _notify(CLUSTER_EVENT_TEMPLATE % ev_type, {
        'cluster_id': cluster_id, 'cluster_name': cluster_name,
        'cluster_status': cluster_status})


def health_notify(cluster, health_check):
    """Sends notification about current cluster health."""
    _notify(HEALTH_EVENT_TYPE,
            _health_notification_body(cluster, health_check))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from unittest import mock


from sahara.conductor import objects as o
from sahara import context
from sahara.tests.unit import base
import sahara.topology.topology_helper as th


class TopologyTestCase(base.SaharaTestCase):
    def setUp(self):
        super(TopologyTestCase, self).setUp()
        context.set_ctx(context.Context(None, None, None, None))

    def test_core_config(self):
        result = th.vm_awareness_core_config()
        self.assertEqual(4, len(result))
        for item in result:
            del item['description']
        className = 'org.apache.hadoop.net.NetworkTopologyWithNodeGroup'
        self.assertIn({'name': "net.topology.impl",
                       'value': className},
                      result)
        self.assertIn({'name': "net.topology.nodegroup.aware",
                       'value': 'true'},
                      result)
        className = ('org.apache.hadoop.hdfs.server.namenode.'
                     'BlockPlacementPolicyWithNodeGroup')
        self.assertIn({'name': "dfs.block.replicator.classname",
                       'value': className},
                      result)

    def test_map_red_config(self):
        result = th.vm_awareness_mapred_config()
        self.assertEqual(3, len(result))
        for item in result:
            del item['description']

        self.assertIn({'name': "mapred.jobtracker.nodegroup.aware",
                       'value': 'true'},
                      result)

        self.assertIn({'name': "mapred.task.cache.levels",
                       'value': '3'},
                      result)
        className = 'org.apache.hadoop.mapred.JobSchedulableWithNodeGroup'
        self.assertIn({'name': "mapred.jobtracker.jobSchedulable",
                       'value': className},
                      result)

    @mock.patch('sahara.utils.openstack.nova.client')
    @mock.patch('sahara.topology.topology_helper._read_compute_topology')
    @mock.patch('sahara.topology.topology_helper._read_swift_topology')
    def test_get_topology(self,
                          swift_topology,
                          compute_topology,
                          novaclient):
        nova = mock.Mock()
        novaclient.return_value = nova
        r1 = mock.Mock()
        r1.hostId = "o1"
        r2 = mock.Mock()
        r2.hostId = "o1"
        r3 = mock.Mock()
        r3.hostId = "o2"
        nova.servers.get.side_effect = [r1, r2, r3, r1, r2, r3]

        swift_topology.return_value = {"s1": "/r1"}
        compute_topology.return_value = {"o1": "/r1", "o2": "/r2"}

        i1 = o.Instance()
        i1.instance_id = "i1"
        i1.instance_name = "i1"
        i1.internal_ip = "0.0.1.1"
        i1.management_ip = "1.1.1.1"

        i2 = o.Instance()
        i2.instance_id = "i2"
        i2.instance_name = "i2"
        i2.management_ip = "1.1.1.2"
        i2.internal_ip = "0.0.1.2"

        i3 = o.Instance()
        i3.instance_id = "i3"
        i3.instance_name = "i3"
        i3.internal_ip = "1.1.1.3"
        i3.management_ip = "0.0.1.3"

        ng1 = o.NodeGroup()
        ng1.name = "1"
        ng1.instances = [i1, i2]

        ng2 = o.NodeGroup()
        ng2.name = "2"
        ng2.instances = [i3]

        cluster = o.Cluster()
        cluster.node_groups = [ng1, ng2]

        top = th.generate_topology_map(cluster, False)
        self.assertEqual({
            "i1": "/r1",
            "1.1.1.1": "/r1",
            "0.0.1.1": "/r1",
            "i2": "/r1",
            "1.1.1.2": "/r1",
            "0.0.1.2": "/r1",
            "i3": "/r2",
            "1.1.1.3": "/r2",
            "0.0.1.3": "/r2",
            "s1": "/r1"
        }, top)

        top = th.generate_topology_map(cluster, True)
        self.assertEqual({
            "i1": "/r1/o1",
            "1.1.1.1": "/r1/o1",
            "0.0.1.1": "/r1/o1",
            "i2": "/r1/o1",
            "1.1.1.2": "/r1/o1",
            "0.0.1.2": "/r1/o1",
            "i3": "/r2/o2",
            "1.1.1.3": "/r2/o2",
            "0.0.1.3": "/r2/o2",
            "s1": "/r1"
        }, top)

    def _read_swift_topology(self, content):
        temp_file = tempfile.NamedTemporaryFile()
        try:
            temp_file.write(str.encode(content))
            temp_file.flush()
            self.override_config("swift_topology_file", temp_file.name)
            return th._read_swift_topology()
        finally:
            temp_file.close()

    def test_read_swift_topology(self):
        topology = self._read_swift_topology("")
        self.assertEqual({}, topology)

        topology = self._read_swift_topology(
            "192.168.1.1 /rack1\n192.168.1.2 /rack2")
        self.assertEqual(
            {"192.168.1.1": "/rack1", "192.168.1.2": "/rack2"}, topology)

        topology = self._read_swift_topology(
            "192.168.1.1 /rack1\n192.168.1.2 /rack2\n\n")
        self.assertEqual(
            {"192.168.1.1": "/rack1", "192.168.1.2": "/rack2"}, topology)

    def _read_compute_topology(self, content):
        temp_file = tempfile.NamedTemporaryFile()
        try:
            temp_file.write(content)
            temp_file.flush()
            self.override_config("compute_topology_file", temp_file.name)
            return th._read_compute_topology()
        finally:
            temp_file.close()

    def test_read_compute_topology(self):
        topology = self._read_swift_topology("")
        self.assertEqual({}, topology)

        topology = self._read_swift_topology(
            "192.168.1.1 /rack1\n192.168.1.2 /rack2")
        self.assertEqual(2, len(topology))

        topology = self._read_swift_topology(
            "192.168.1.1 /rack1\n192.168.1.2 /rack2\n\n")
        self.assertEqual(2, len(topology))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_utils import excutils
import six

from sahara import conductor as c
from sahara import context
from sahara.plugins import base as plugin_base
from sahara.service import api
from sahara.service.health import verification_base
from sahara.service import quotas
from sahara.utils import cluster as c_u
from sahara.utils import general as g
from sahara.utils.notification import sender
from sahara.utils.openstack import base as b
from sahara.utils.openstack import images as sahara_images


conductor = c.API
CONF = cfg.CONF


# Cluster ops

def get_clusters(**kwargs):
    return conductor.cluster_get_all(context.ctx(),
                                     regex_search=True, **kwargs)


def get_cluster(id, show_progress=False):
    return conductor.cluster_get(context.ctx(), id, show_progress)


def scale_cluster(id, data):
    context.set_current_cluster_id(id)
    ctx = context.ctx()

    cluster = conductor.cluster_get(ctx, id)
    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)
    existing_node_groups = data.get('resize_node_groups', [])
    additional_node_groups = data.get('add_node_groups', [])

    # the next map is the main object we will work with
    # to_be_enlarged : {node_group_id: desired_amount_of_instances}
    to_be_enlarged = {}
    for ng in existing_node_groups:
        ng_id = g.find(cluster.node_groups, name=ng['name'])['id']
        to_be_enlarged.update({ng_id: ng['count']})

    additional = construct_ngs_for_scaling(cluster, additional_node_groups)
    cluster = conductor.cluster_get(ctx, cluster)
    _add_ports_for_auto_sg(ctx, cluster, plugin)

    try:
        cluster = c_u.change_cluster_status(
            cluster, c_u.CLUSTER_STATUS_VALIDATING)
        quotas.check_scaling(cluster, to_be_enlarged, additional)
        plugin.recommend_configs(cluster, scaling=True)
        plugin.validate_scaling(cluster, to_be_enlarged, additional)
    except Exception as e:
        with excutils.save_and_reraise_exception():
            c_u.clean_cluster_from_empty_ng(cluster)
            c_u.change_cluster_status(
                cluster, c_u.CLUSTER_STATUS_ACTIVE, six.text_type(e))

    # If we are here validation is successful.
    # So let's update to_be_enlarged map:
    to_be_enlarged.update(additional)

    for node_group in cluster.node_groups:
        if node_group.id not in to_be_enlarged:
            to_be_enlarged[node_group.id] = node_group.count

    api.OPS.provision_scaled_cluster(id, to_be_enlarged)
    return cluster


def create_cluster(values):
    plugin = plugin_base.PLUGINS.get_plugin(values['plugin_name'])
    return _cluster_create(values, plugin)


def create_multiple_clusters(values):
    num_of_clusters = values['count']
    clusters = []
    plugin = plugin_base.PLUGINS.get_plugin(values['plugin_name'])
    for counter in range(num_of_clusters):
        cluster_dict = values.copy()
        cluster_name = cluster_dict['name']
        cluster_dict['name'] = get_multiple_cluster_name(num_of_clusters,
                                                         cluster_name,
                                                         counter + 1)
        cluster = _cluster_create(cluster_dict, plugin)

        clusters.append(cluster.id)

    clusters_dict = {'clusters': clusters}
    return clusters_dict


def _cluster_create(values, plugin):
    ctx = context.ctx()
    cluster = conductor.cluster_create(ctx, values)
    context.set_current_cluster_id(cluster.id)
    sender.status_notify(cluster.id, cluster.name, "New",
                         "create")
    _add_ports_for_auto_sg(ctx, cluster, plugin)

    # validating cluster
    try:
        plugin.recommend_configs(cluster)
        cluster = c_u.change_cluster_status(
            cluster, c_u.CLUSTER_STATUS_VALIDATING)
        plugin.validate(cluster)
        quotas.check_cluster(cluster)
    except Exception as e:
        with excutils.save_and_reraise_exception():
            c_u.change_cluster_status(
                cluster, c_u.CLUSTER_STATUS_ERROR, six.text_type(e))

    api.OPS.provision_cluster(cluster.id)

    return cluster


def get_multiple_cluster_name(num_of_clusters, name, counter):
    return "%%s-%%0%dd" % len(str(num_of_clusters)) % (name, counter)


def _add_ports_for_auto_sg(ctx, cluster, plugin):
    for ng in cluster.node_groups:
        if ng.auto_security_group:
            ports = {'open_ports': plugin.get_open_ports(ng)}
            conductor.node_group_update(ctx, ng, ports)


def terminate_cluster(id):
    context.set_current_cluster_id(id)
    cluster = c_u.change_cluster_status(id, c_u.CLUSTER_STATUS_DELETING)

    if cluster is None:
        return

    api.OPS.terminate_cluster(id)
    sender.status_notify(cluster.id, cluster.name, cluster.status,
                         "delete")


def update_cluster(id, values):
    if verification_base.update_verification_required(values):
        api.OPS.handle_verification(id, values)
        return conductor.cluster_get(context.ctx(), id)
    return conductor.cluster_update(context.ctx(), id, values)


# ClusterTemplate ops

def get_cluster_templates(**kwargs):
    return conductor.cluster_template_get_all(context.ctx(),
                                              regex_search=True, **kwargs)


def get_cluster_template(id):
    return conductor.cluster_template_get(context.ctx(), id)


def create_cluster_template(values):
    return conductor.cluster_template_create(context.ctx(), values)


def terminate_cluster_template(id):
    return conductor.cluster_template_destroy(context.ctx(), id)


def update_cluster_template(id, values):
    return conductor.cluster_template_update(context.ctx(), id, values)


# NodeGroupTemplate ops

def get_node_group_templates(**kwargs):
    return conductor.node_group_template_get_all(context.ctx(),
                                                 regex_search=True, **kwargs)


def get_node_group_template(id):
    return conductor.node_group_template_get(context.ctx(), id)


def create_node_group_template(values):
    return conductor.node_group_template_create(context.ctx(), values)


def terminate_node_group_template(id):
    return conductor.node_group_template_destroy(context.ctx(), id)


def update_node_group_template(id, values):
    return conductor.node_group_template_update(context.ctx(), id, values)


def export_node_group_template(id):
    return conductor.node_group_template_get(context.ctx(), id)


# Plugins ops

def get_plugins():
    return plugin_base.PLUGINS.get_plugins(serialized=True)


def get_plugin(plugin_name, version=None):
    return plugin_base.PLUGINS.serialize_plugin(plugin_name, version)


def update_plugin(plugin_name, values):
    return plugin_base.PLUGINS.update_plugin(plugin_name, values)


def construct_ngs_for_scaling(cluster, additional_node_groups):
    ctx = context.ctx()
    additional = {}
    for ng in additional_node_groups:
        count = ng['count']
        ng['count'] = 0
        ng_id = conductor.node_group_add(ctx, cluster, ng)
        additional.update({ng_id: count})
    return additional

# Image Registry


def get_images(name, tags):
    return b.execute_with_retries(
        sahara_images.image_manager().list_registered, name, tags)


def get_image(**kwargs):
    if len(kwargs) == 1 and 'id' in kwargs:
        return b.execute_with_retries(
            sahara_images.image_manager().get, kwargs['id'])
    else:
        return b.execute_with_retries(
            sahara_images.image_manager().find, **kwargs)


def get_registered_image(image_id):
    return b.execute_with_retries(
        sahara_images.image_manager().get_registered_image, image_id)


def register_image(image_id, username, description=None):
    manager = sahara_images.image_manager()
    b.execute_with_retries(
        manager.set_image_info, image_id, username, description)
    return b.execute_with_retries(manager.get, image_id)


def unregister_image(image_id):
    manager = sahara_images.image_manager()
    b.execute_with_retries(manager.unset_image_info, image_id)
    return b.execute_with_retries(manager.get, image_id)


def add_image_tags(image_id, tags):
    manager = sahara_images.image_manager()
    b.execute_with_retries(manager.tag, image_id, tags)
    return b.execute_with_retries(manager.get, image_id)


def remove_image_tags(image_id, tags):
    manager = sahara_images.image_manager()
    b.execute_with_retries(manager.untag, image_id, tags)
    return b.execute_with_retries(manager.get, image_id)



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from sahara.api import acl
from sahara.service.api.v2 import clusters as api
from sahara.service import validation as v
from sahara.service.validations import clusters as v_c
from sahara.service.validations import clusters_scaling as v_c_s
from sahara.service.validations import clusters_schema as v_c_schema
import sahara.utils.api as u


rest = u.RestV2('clusters', __name__)


def _replace_tenant_id_project_id_provision_steps(c):
    if 'provision_progress' in c:
        for step in c['provision_progress']:
            dict.update(step, {'project_id': step['tenant_id']})
            dict.pop(step, 'tenant_id')


@rest.get('/clusters')
@acl.enforce("data-processing:cluster:list")
@v.check_exists(api.get_cluster, 'marker')
@v.validate(None, v.validate_pagination_limit)
@v.validate_request_params(['plugin_name', 'plugin_version', 'name'])
def clusters_list():
    request_args = u.get_request_args().to_dict()
    if 'plugin_version' in request_args:
        request_args['hadoop_version'] = request_args['plugin_version']
        del request_args['plugin_version']
    result = api.get_clusters(**request_args)
    for c in result:
        u._replace_hadoop_version_plugin_version(c)
        u._replace_tenant_id_project_id(c)
        _replace_tenant_id_project_id_provision_steps(c)
    return u.render(res=result, name='clusters')


@rest.post('/clusters')
@acl.enforce("data-processing:cluster:create")
@v.validate(v_c_schema.CLUSTER_SCHEMA_V2,
            v_c.check_one_or_multiple_clusters_create)
@v.validate_request_params([])
def clusters_create(data):
    # renaming hadoop_version -> plugin_version
    # this can be removed once APIv1 is deprecated
    data['hadoop_version'] = data['plugin_version']
    del data['plugin_version']
    if data.get('count', None) is not None:
        result = api.create_multiple_clusters(data)
        for c in result['clusters']:
            u._replace_hadoop_version_plugin_version(c['cluster'])
            u._replace_tenant_id_project_id(c['cluster'])
        return u.render(result)
    else:
        result = api.create_cluster(data).to_wrapped_dict()
        u._replace_hadoop_version_plugin_version(result['cluster'])
        u._replace_tenant_id_project_id(result['cluster'])
        return u.render(result)


@rest.put('/clusters/<cluster_id>')
@acl.enforce("data-processing:cluster:scale")
@v.check_exists(api.get_cluster, 'cluster_id')
@v.validate(v_c_schema.CLUSTER_SCALING_SCHEMA_V2, v_c_s.check_cluster_scaling)
@v.validate_request_params([])
def clusters_scale(cluster_id, data):
    result = u.to_wrapped_dict_no_render(
        api.scale_cluster, cluster_id, data)
    u._replace_hadoop_version_plugin_version(result['cluster'])
    u._replace_tenant_id_project_id(result['cluster'])
    _replace_tenant_id_project_id_provision_steps(result['cluster'])
    return u.render(result)


@rest.get('/clusters/<cluster_id>')
@acl.enforce("data-processing:cluster:get")
@v.check_exists(api.get_cluster, 'cluster_id')
@v.validate_request_params(['show_progress'])
def clusters_get(cluster_id):
    data = u.get_request_args()
    show_events = six.text_type(
        data.get('show_progress', 'false')).lower() == 'true'
    result = u.to_wrapped_dict_no_render(
        api.get_cluster, cluster_id, show_events)
    u._replace_hadoop_version_plugin_version(result['cluster'])
    u._replace_tenant_id_project_id(result['cluster'])
    _replace_tenant_id_project_id_provision_steps(result['cluster'])
    return u.render(result)


@rest.patch('/clusters/<cluster_id>')
@acl.enforce("data-processing:cluster:update")
@v.check_exists(api.get_cluster, 'cluster_id')
@v.validate(v_c_schema.CLUSTER_UPDATE_SCHEMA_V2, v_c.check_cluster_update)
@v.validate_request_params([])
def clusters_update(cluster_id, data):
    result = u.to_wrapped_dict_no_render(
        api.update_cluster, cluster_id, data)
    u._replace_hadoop_version_plugin_version(result['cluster'])
    u._replace_tenant_id_project_id(result['cluster'])
    _replace_tenant_id_project_id_provision_steps(result['cluster'])
    return u.render(result)


@rest.delete('/clusters/<cluster_id>')
@acl.enforce("data-processing:cluster:delete")
@v.check_exists(api.get_cluster, 'cluster_id')
@v.validate(v_c_schema.CLUSTER_DELETE_SCHEMA_V2, v_c.check_cluster_delete)
@v.validate_request_params([])
def clusters_delete(cluster_id):
    data = u.request_data()
    force = data.get('force', False)
    extra = api.get_cluster(cluster_id).get('extra', {})
    stack_name = extra.get('heat_stack_name', None) if extra else None
    api.terminate_cluster(cluster_id, force=force)
    if force:
        return u.render({"stack_name": stack_name}, status=200)
    else:
        return u.render(res=None, status=204)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import mock

from sqlalchemy import exc as sa_exc
import testtools

from sahara.conductor import manager
from sahara import context
from sahara.db.sqlalchemy import models as m
from sahara import exceptions as ex
import sahara.tests.unit.conductor.base as test_base
from sahara.utils import cluster as c_u


SAMPLE_CLUSTER = {
    "plugin_name": "test_plugin",
    "hadoop_version": "test_version",
    "tenant_id": "tenant_1",
    "name": "test_cluster",
    "user_keypair_id": "my_keypair",
    "node_groups": [
        {
            "name": "ng_1",
            "flavor_id": "42",
            "node_processes": ["p1", "p2"],
            "count": 1,
            "security_groups": None,
            'use_autoconfig': True,
            "shares": None
        },
        {
            "name": "ng_2",
            "flavor_id": "42",
            "node_processes": ["p3", "p4"],
            "count": 3,
            "security_groups": ["group1", "group2"],
            'use_autoconfig': True,
            "shares": None
        }
    ],
    "cluster_configs": {
        "service_1": {
            "config_2": "value_2"
        },
        "service_2": {
            "config_1": "value_1"
        }
    },
    "shares": [],
    "is_public": False,
    "is_protected": False
}


class ClusterTest(test_base.ConductorManagerTestCase):
    def __init__(self, *args, **kwargs):
        super(ClusterTest, self).__init__(
            checks=[
                lambda: SAMPLE_CLUSTER,
                lambda: manager.CLUSTER_DEFAULTS,
                lambda: manager.NODE_GROUP_DEFAULTS,
                lambda: manager.INSTANCE_DEFAULTS,
            ], *args, **kwargs)

    def test_cluster_create_list_update_delete(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        self.assertIsInstance(cluster_db_obj, dict)

        lst = self.api.cluster_get_all(ctx)
        self.assertEqual(1, len(lst))
        cl_id = lst[0]["id"]

        updated_cl = self.api.cluster_update(
            ctx, cl_id, {"is_public": True})
        self.assertIsInstance(updated_cl, dict)
        self.assertEqual(True, updated_cl["is_public"])

        self.api.cluster_destroy(ctx, cl_id)
        lst = self.api.cluster_get_all(ctx)
        self.assertEqual(0, len(lst))

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.cluster_destroy(ctx, cl_id)

    def test_duplicate_cluster_create(self):
        ctx = context.ctx()
        self.api.cluster_create(ctx, SAMPLE_CLUSTER)

        with testtools.ExpectedException(ex.DBDuplicateEntry):
            self.api.cluster_create(ctx, SAMPLE_CLUSTER)

    def test_cluster_fields(self):
        ctx = context.ctx()
        cl_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        self.assertIsInstance(cl_db_obj, dict)

        for key, val in SAMPLE_CLUSTER.items():
            if key == 'node_groups':
                # this will be checked separately
                continue
            self.assertEqual(val, cl_db_obj.get(key),
                             "Key not found %s" % key)

        for ng in cl_db_obj["node_groups"]:
            ng.pop("created_at")
            ng.pop("updated_at")
            ng.pop("id")
            self.assertEqual(cl_db_obj["id"], ng.pop("cluster_id"))
            ng.pop("image_id")
            self.assertEqual([], ng.pop("instances"))
            ng.pop("node_configs")
            ng.pop("node_group_template_id")
            ng.pop("volume_mount_prefix")
            ng.pop("volumes_size")
            ng.pop("volumes_per_node")
            ng.pop("volumes_availability_zone")
            ng.pop("volume_type")
            ng.pop("floating_ip_pool")
            ng.pop("boot_from_volume")
            ng.pop("boot_volume_type")
            ng.pop("boot_volume_availability_zone")
            ng.pop("boot_volume_local_to_instance")
            ng.pop("image_username")
            ng.pop("open_ports")
            ng.pop("auto_security_group")
            ng.pop("is_proxy_gateway")
            ng.pop("tenant_id")
            ng.pop("availability_zone")
            ng.pop('volume_local_to_instance')

        self.assertEqual(SAMPLE_CLUSTER["node_groups"],
                         cl_db_obj["node_groups"])

    def test_cluster_no_ng(self):
        ctx = context.ctx()
        cluster_schema = copy.deepcopy(SAMPLE_CLUSTER)
        cluster_schema.pop('node_groups')
        cl_db_obj = self.api.cluster_create(ctx, cluster_schema)
        self.assertIsInstance(cl_db_obj, dict)

        for key, val in cluster_schema.items():
            self.assertEqual(val, cl_db_obj.get(key),
                             "Key not found %s" % key)

        self.assertEqual([], cl_db_obj["node_groups"])

    def test_cluster_update_status(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        updated_cl = self.api.cluster_update(
            ctx, _id, {"status": c_u.CLUSTER_STATUS_ACTIVE})
        self.assertIsInstance(updated_cl, dict)
        self.assertEqual(c_u.CLUSTER_STATUS_ACTIVE, updated_cl["status"])

        get_cl_obj = self.api.cluster_get(ctx, _id)
        self.assertEqual(updated_cl, get_cl_obj)

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.cluster_update(
                ctx, "bad_id", {"status": c_u.CLUSTER_STATUS_ACTIVE})

    def _ng_in_cluster(self, cluster_db_obj, ng_id):
        for ng in cluster_db_obj["node_groups"]:
            if ng["id"] == ng_id:
                return ng
        return None

    def test_add_node_group(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        node_group = {
            "name": "ng_3",
            "flavor_id": "42",
            "node_processes": ["p3", "p4"],
            "count": 5
        }

        ng_id = self.api.node_group_add(ctx, _id, node_group)

        cluster_db_obj = self.api.cluster_get(ctx, _id)
        found_ng = self._ng_in_cluster(cluster_db_obj, ng_id)

        self.assertTrue(found_ng, "New Node Group not found")

    def test_update_node_group(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        self.assertEqual(2, len(cluster_db_obj["node_groups"]))
        ng_id = cluster_db_obj["node_groups"][-1]["id"]

        self.api.node_group_update(ctx, ng_id, {"image_id": "test_image"})

        cluster_db_obj = self.api.cluster_get(ctx, _id)

        found_ng = self._ng_in_cluster(cluster_db_obj, ng_id)
        self.assertTrue(found_ng, "Updated Node Group not found")

        for ng in cluster_db_obj["node_groups"]:
            if ng["id"] != ng_id:
                continue

            self.assertEqual("test_image", ng["image_id"])

    def test_delete_node_group(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        ng_id = cluster_db_obj["node_groups"][-1]["id"]

        self.api.node_group_remove(ctx, ng_id)

        cluster_db_obj = self.api.cluster_get(ctx, _id)
        found_ng = self._ng_in_cluster(cluster_db_obj, ng_id)

        self.assertFalse(found_ng, "Node Group is still in a CLuster")

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.node_group_remove(ctx, ng_id)

    def _add_instance(self, ctx, ng_id):
        instance = {
            "instance_name": "additional_vm"
        }
        return self.api.instance_add(ctx, ng_id, instance)

    def _add_instance_ipv6(self, ctx, ng_id, instance_name):
        instance = {
            "instance_name": instance_name,
            "internal_ip": "FE80:0000:0000:0000:0202:B3FF:FE1E:8329",
            "management_ip": "FE80:0000:0000:0000:0202:B3FF:FE1E:8329"
        }
        return self.api.instance_add(ctx, ng_id, instance)

    def test_add_instance(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        ng_id = cluster_db_obj["node_groups"][-1]["id"]
        count = cluster_db_obj["node_groups"][-1]["count"]

        self._add_instance(ctx, ng_id)

        cluster_db_obj = self.api.cluster_get(ctx, _id)
        for ng in cluster_db_obj["node_groups"]:
            if ng["id"] != ng_id:
                continue

            ng.pop('tenant_id')
            self.assertEqual(count + 1, ng["count"])
            self.assertEqual("additional_vm",
                             ng["instances"][0]["instance_name"])

    def test_add_instance_ipv6(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        ng_id = cluster_db_obj["node_groups"][-1]["id"]
        count = cluster_db_obj["node_groups"][-1]["count"]

        instance_name = "additional_vm_ipv6"
        self._add_instance_ipv6(ctx, ng_id, instance_name)

        cluster_db_obj = self.api.cluster_get(ctx, _id)
        for ng in cluster_db_obj["node_groups"]:
            if ng["id"] != ng_id:
                continue

            ng.pop('tenant_id')
            self.assertEqual(count + 1, ng["count"])
            self.assertEqual(instance_name,
                             ng["instances"][0]["instance_name"])

    def test_update_instance(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        ng_id = cluster_db_obj["node_groups"][-1]["id"]

        instance_id = self._add_instance(ctx, ng_id)

        self.api.instance_update(ctx, instance_id,
                                 {"management_ip": "1.1.1.1"})

        cluster_db_obj = self.api.cluster_get(ctx, _id)
        for ng in cluster_db_obj["node_groups"]:
            if ng["id"] != ng_id:
                continue

            self.assertEqual("1.1.1.1", ng["instances"][0]["management_ip"])

    def test_update_instance_ipv6(self):
        ctx = context.ctx()
        ip = "FE80:0000:0000:0000:0202:B3FF:FE1E:8329"
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        ng_id = cluster_db_obj["node_groups"][-1]["id"]

        instance_id = self._add_instance(ctx, ng_id)

        self.api.instance_update(ctx, instance_id, {"management_ip": ip})

        cluster_db_obj = self.api.cluster_get(ctx, _id)
        for ng in cluster_db_obj["node_groups"]:
            if ng["id"] != ng_id:
                continue

            self.assertEqual(ip, ng["instances"][0]["management_ip"])

    def test_remove_instance(self):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        ng_id = cluster_db_obj["node_groups"][-1]["id"]
        count = cluster_db_obj["node_groups"][-1]["count"]

        instance_id = self._add_instance(ctx, ng_id)

        cluster_db_obj = self.api.cluster_get(ctx, _id)
        for ng in cluster_db_obj["node_groups"]:
            if ng["id"] != ng_id:
                continue

            self.assertEqual(count + 1, ng["count"])

        self.api.instance_remove(ctx, instance_id)

        cluster_db_obj = self.api.cluster_get(ctx, _id)
        for ng in cluster_db_obj["node_groups"]:
            if ng["id"] != ng_id:
                continue

            self.assertEqual(count, ng["count"])

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.instance_remove(ctx, instance_id)

    def test_cluster_search(self):
        ctx = context.ctx()
        vals = copy.deepcopy(SAMPLE_CLUSTER)
        vals['name'] = "test_name"
        self.api.cluster_create(ctx, vals)

        lst = self.api.cluster_get_all(ctx)
        self.assertEqual(1, len(lst))

        kwargs = {'name': vals['name'],
                  'plugin_name': vals['plugin_name']}
        lst = self.api.cluster_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Valid field but no matching value
        kwargs = {'name': vals['name']+'foo'}
        lst = self.api.cluster_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Valid field with substrings
        kwargs = {'name': 'test'}
        lst = self.api.cluster_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Invalid field
        self.assertRaises(sa_exc.InvalidRequestError,
                          self.api.cluster_get_all,
                          ctx, **{'badfield': 'somevalue'})

    @mock.patch('sahara.db.sqlalchemy.api.regex_filter')
    def test_cluster_search_regex(self, regex_filter):

        # do this so we can return the correct value
        def _regex_filter(query, cls, regex_cols, search_opts):
            return query, search_opts

        regex_filter.side_effect = _regex_filter

        ctx = context.ctx()
        self.api.cluster_get_all(ctx)
        self.assertEqual(0, regex_filter.call_count)

        self.api.cluster_get_all(ctx, regex_search=True, name="fox")
        self.assertEqual(1, regex_filter.call_count)
        args, kwargs = regex_filter.call_args
        self.assertIs(args[1], m.Cluster)
        self.assertEqual(args[2], ["name", "description", "plugin_name",
                                   "tenant_id"])
        self.assertEqual(args[3], {"name": "fox"})

    @mock.patch("sahara.service.edp.utils.shares.mount_shares")
    def test_cluster_update_shares(self, mount_shares):
        ctx = context.ctx()
        cluster_db_obj = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        _id = cluster_db_obj["id"]

        test_shares = [
            {
                "id": "bd71d2d5-60a0-4ed9-a3d2-ad312c368880",
                "path": "/mnt/manila",
                "access_level": "rw"
            }
        ]

        updated_cl = self.api.cluster_update(ctx, _id, {"shares": test_shares})
        self.assertIsInstance(updated_cl, dict)
        self.assertEqual(test_shares, updated_cl["shares"])

        get_cl_obj = self.api.cluster_get(ctx, _id)
        self.assertEqual(updated_cl, get_cl_obj)



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from castellan.common.objects import passphrase as key
from castellan.key_manager import key_manager as km


"""sahara.service.castellan.sahara_key_manager

This module contains the KeyManager class that will be used by the
castellan library, it is not meant for direct usage within sahara.
"""


class SaharaKeyManager(km.KeyManager):
    """Sahara specific key manager

    This manager is a thin wrapper around the secret being stored. It is
    intended for backward compatible use only. It will not store keys
    or generate UUIDs but instead return the secret that is being stored.
    This behavior allows Sahara to continue storing secrets in its database
    while using the Castellan key manager abstraction.
    """
    def __init__(self, configuration=None):
        pass

    def create_key(self, context, algorithm=None, length=0,
                   expiration=None, **kwargs):
        """creates a key

        algorithm, length, and expiration are unused by sahara keys.
        """
        return key.Passphrase(passphrase=kwargs.get('passphrase', ''))

    def create_key_pair(self, *args, **kwargs):
        pass

    def store(self, context, key, expiration=None, **kwargs):
        """store a key

        in normal usage a store_key will return the UUID of the key as
        dictated by the key manager. Sahara would then store this UUID in
        its database to use for retrieval. As sahara is not actually using
        a key manager in this context it will return the key's payload for
        storage.
        """
        return key.get_encoded()

    def get(self, context, key_id, **kwargs):
        """get a key

        since sahara is not actually storing key UUIDs the key_id to this
        function should actually be the key payload. this function will
        simply return a new SaharaKey based on that value.
        """
        return key.Passphrase(passphrase=key_id)

    def delete(self, context, key_id, **kwargs):
        """delete a key

        as there is no external key manager, this function will not
        perform any external actions. therefore, it won't change anything.
        """
        pass

    def list(self, *args, **kwargs):
        """list all managed keys

        current implementation of the key manager does not utilize this
        """
        pass



# Copyright (c) 2017 EasyStack Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.plugins import base as pl_base
from sahara.plugins import provisioning as pr_base
from sahara.service.api.v2 import plugins as api
from sahara.tests.unit import base
import sahara.tests.unit.service.api.v2.base as api_base


class TestPluginApi(base.SaharaWithDbTestCase):
    def setUp(self):
        super(TestPluginApi, self).setUp()
        self.calls_order = []
        self.override_config('plugins', ['fake'])
        pl_base.PLUGINS = api_base.FakePluginManager(self.calls_order)

    def test_get_plugin(self):
        # processing to dict
        data = api.get_plugin('fake', '0.1').dict
        self.assertIsNotNone(data)
        self.assertEqual(
            len(pr_base.list_of_common_configs()), len(data.get('configs')))
        self.assertEqual(['fake', '0.1'], data.get('required_image_tags'))
        self.assertEqual(
            {'HDFS': ['namenode', 'datanode']}, data.get('node_processes'))

        self.assertIsNone(api.get_plugin('fake', '0.3'))
        data = api.get_plugin('fake').dict
        self.assertIsNotNone(data.get('version_labels'))
        self.assertIsNotNone(data.get('plugin_labels'))
        del data['plugin_labels']
        del data['version_labels']

        self.assertEqual({
            'description': "Some description",
            'name': 'fake',
            'title': 'Fake plugin',
            'versions': ['0.1', '0.2']}, data)
        self.assertIsNone(api.get_plugin('name1', '0.1'))

    def test_update_plugin(self):
        data = api.get_plugin('fake', '0.1').dict
        self.assertIsNotNone(data)

        updated = api.update_plugin('fake', values={
            'plugin_labels': {'enabled': {'status': False}}}).dict
        self.assertFalse(updated['plugin_labels']['enabled']['status'])

        updated = api.update_plugin('fake', values={
            'plugin_labels': {'enabled': {'status': True}}}).dict
        self.assertTrue(updated['plugin_labels']['enabled']['status'])

        # restore to original status
        updated = api.update_plugin('fake', values={
            'plugin_labels': data['plugin_labels']}).dict
        self.assertEqual(data['plugin_labels']['enabled']['status'],
                         updated['plugin_labels']['enabled']['status'])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from sahara.api import acl
from sahara.service.api import v10 as api
from sahara.service import validation as v
from sahara.service.validations import cluster_template_schema as ct_schema
from sahara.service.validations import cluster_templates as v_ct
from sahara.service.validations import clusters as v_c
from sahara.service.validations import clusters_scaling as v_c_s
from sahara.service.validations import clusters_schema as v_c_schema
from sahara.service.validations import images as v_images
from sahara.service.validations import node_group_template_schema as ngt_schema
from sahara.service.validations import node_group_templates as v_ngt
from sahara.service.validations import plugins as v_p
import sahara.utils.api as u


rest = u.Rest('v10', __name__)


# Cluster ops

@rest.get('/clusters')
@acl.enforce("data-processing:clusters:get_all")
@v.check_exists(api.get_cluster, 'marker')
@v.validate(None, v.validate_pagination_limit,
            v.validate_sorting_clusters)
def clusters_list():
    result = api.get_clusters(**u.get_request_args().to_dict())
    return u.render(res=result, name='clusters')


@rest.post('/clusters')
@acl.enforce("data-processing:clusters:create")
@v.validate(v_c_schema.CLUSTER_SCHEMA, v_c.check_cluster_create)
def clusters_create(data):
    return u.render(api.create_cluster(data).to_wrapped_dict())


@rest.post('/clusters/multiple')
@acl.enforce("data-processing:clusters:create")
@v.validate(
    v_c_schema.MULTIPLE_CLUSTER_SCHEMA, v_c.check_multiple_clusters_create)
def clusters_create_multiple(data):
    return u.render(api.create_multiple_clusters(data))


@rest.put('/clusters/<cluster_id>')
@acl.enforce("data-processing:clusters:scale")
@v.check_exists(api.get_cluster, 'cluster_id')
@v.validate(v_c_schema.CLUSTER_SCALING_SCHEMA, v_c_s.check_cluster_scaling)
def clusters_scale(cluster_id, data):
    return u.to_wrapped_dict(api.scale_cluster, cluster_id, data)


@rest.get('/clusters/<cluster_id>')
@acl.enforce("data-processing:clusters:get")
@v.check_exists(api.get_cluster, 'cluster_id')
def clusters_get(cluster_id):
    data = u.get_request_args()
    show_events = six.text_type(
        data.get('show_progress', 'false')).lower() == 'true'
    return u.to_wrapped_dict(api.get_cluster, cluster_id, show_events)


@rest.patch('/clusters/<cluster_id>')
@acl.enforce("data-processing:clusters:modify")
@v.check_exists(api.get_cluster, 'cluster_id')
@v.validate(v_c_schema.CLUSTER_UPDATE_SCHEMA, v_c.check_cluster_update)
def clusters_update(cluster_id, data):
    return u.to_wrapped_dict(api.update_cluster, cluster_id, data)


@rest.delete('/clusters/<cluster_id>')
@acl.enforce("data-processing:clusters:delete")
@v.check_exists(api.get_cluster, 'cluster_id')
@v.validate(None, v_c.check_cluster_delete)
def clusters_delete(cluster_id):
    api.terminate_cluster(cluster_id)
    return u.render()


# ClusterTemplate ops

@rest.get('/cluster-templates')
@acl.enforce("data-processing:cluster-templates:get_all")
@v.check_exists(api.get_cluster_template, 'marker')
@v.validate(None, v.validate_pagination_limit,
            v.validate_sorting_cluster_templates)
def cluster_templates_list():
    result = api.get_cluster_templates(
        **u.get_request_args().to_dict())

    return u.render(res=result, name='cluster_templates')


@rest.post('/cluster-templates')
@acl.enforce("data-processing:cluster-templates:create")
@v.validate(ct_schema.CLUSTER_TEMPLATE_SCHEMA,
            v_ct.check_cluster_template_create)
def cluster_templates_create(data):
    return u.render(api.create_cluster_template(data).to_wrapped_dict())


@rest.get('/cluster-templates/<cluster_template_id>')
@acl.enforce("data-processing:cluster-templates:get")
@v.check_exists(api.get_cluster_template, 'cluster_template_id')
def cluster_templates_get(cluster_template_id):
    return u.to_wrapped_dict(api.get_cluster_template, cluster_template_id)


@rest.put('/cluster-templates/<cluster_template_id>')
@acl.enforce("data-processing:cluster-templates:modify")
@v.check_exists(api.get_cluster_template, 'cluster_template_id')
@v.validate(ct_schema.CLUSTER_TEMPLATE_UPDATE_SCHEMA,
            v_ct.check_cluster_template_update)
def cluster_templates_update(cluster_template_id, data):
    return u.to_wrapped_dict(
        api.update_cluster_template, cluster_template_id, data)


@rest.delete('/cluster-templates/<cluster_template_id>')
@acl.enforce("data-processing:cluster-templates:delete")
@v.check_exists(api.get_cluster_template, 'cluster_template_id')
@v.validate(None, v_ct.check_cluster_template_usage)
def cluster_templates_delete(cluster_template_id):
    api.terminate_cluster_template(cluster_template_id)
    return u.render()


def _cluster_template_export_helper(template):
    template.pop('id')
    template.pop('updated_at')
    template.pop('created_at')
    template.pop('tenant_id')
    template.pop('is_default')
    template['default_image_id'] = '{default_image_id}'
    template['node_groups'] = '{node_groups}'


@rest.get('/cluster-templates/<cluster_template_id>/export')
@acl.enforce("data-processing:cluster-templates:get")
@v.check_exists(api.get_cluster_template, 'cluster_template_id')
def cluster_template_export(cluster_template_id):
    content = u.to_wrapped_dict_no_render(
        api.get_cluster_template, cluster_template_id)
    _cluster_template_export_helper(content['cluster_template'])
    res = u.render(content)
    res.headers.add('Content-Disposition', 'attachment',
                    filename='cluster_template.json')
    return res


# NodeGroupTemplate ops

@rest.get('/node-group-templates')
@acl.enforce("data-processing:node-group-templates:get_all")
@v.check_exists(api.get_node_group_template, 'marker')
@v.validate(None, v.validate_pagination_limit,
            v.validate_sorting_node_group_templates)
def node_group_templates_list():
    result = api.get_node_group_templates(
        **u.get_request_args().to_dict())
    return u.render(res=result, name='node_group_templates')


@rest.post('/node-group-templates')
@acl.enforce("data-processing:node-group-templates:create")
@v.validate(ngt_schema.NODE_GROUP_TEMPLATE_SCHEMA,
            v_ngt.check_node_group_template_create)
def node_group_templates_create(data):
    return u.render(api.create_node_group_template(data).to_wrapped_dict())


@rest.get('/node-group-templates/<node_group_template_id>')
@acl.enforce("data-processing:node-group-templates:get")
@v.check_exists(api.get_node_group_template, 'node_group_template_id')
def node_group_templates_get(node_group_template_id):
    return u.to_wrapped_dict(
        api.get_node_group_template, node_group_template_id)


@rest.put('/node-group-templates/<node_group_template_id>')
@acl.enforce("data-processing:node-group-templates:modify")
@v.check_exists(api.get_node_group_template, 'node_group_template_id')
@v.validate(ngt_schema.NODE_GROUP_TEMPLATE_UPDATE_SCHEMA,
            v_ngt.check_node_group_template_update)
def node_group_templates_update(node_group_template_id, data):
    return u.to_wrapped_dict(
        api.update_node_group_template, node_group_template_id, data)


@rest.delete('/node-group-templates/<node_group_template_id>')
@acl.enforce("data-processing:node-group-templates:delete")
@v.check_exists(api.get_node_group_template, 'node_group_template_id')
@v.validate(None, v_ngt.check_node_group_template_usage)
def node_group_templates_delete(node_group_template_id):
    api.terminate_node_group_template(node_group_template_id)
    return u.render()


def _node_group_template_export_helper(template):
    template.pop('id')
    template.pop('updated_at')
    template.pop('created_at')
    template.pop('tenant_id')
    template.pop('is_default')
    template['flavor_id'] = '{flavor_id}'
    template['security_groups'] = '{security_groups}'
    template['image_id'] = '{image_id}'
    template['floating_ip_pool'] = '{floating_ip_pool}'


@rest.get('/node-group-templates/<node_group_template_id>/export')
@acl.enforce("data-processing:node-group-templates:get")
@v.check_exists(api.get_node_group_template, 'node_group_template_id')
def node_group_template_export(node_group_template_id):
    content = u.to_wrapped_dict_no_render(
        api.export_node_group_template, node_group_template_id)
    _node_group_template_export_helper(content['node_group_template'])
    res = u.render(content)
    res.headers.add('Content-Disposition', 'attachment',
                    filename='node_group_template.json')
    return res


# Plugins ops

@rest.get('/plugins')
@acl.enforce("data-processing:plugins:get_all")
def plugins_list():
    return u.render(plugins=[p.dict for p in api.get_plugins()])


@rest.get('/plugins/<plugin_name>')
@acl.enforce("data-processing:plugins:get")
@v.check_exists(api.get_plugin, plugin_name='plugin_name')
def plugins_get(plugin_name):
    return u.render(api.get_plugin(plugin_name).wrapped_dict)


@rest.get('/plugins/<plugin_name>/<version>')
@acl.enforce("data-processing:plugins:get_version")
@v.check_exists(api.get_plugin, plugin_name='plugin_name', version='version')
def plugins_get_version(plugin_name, version):
    return u.render(api.get_plugin(plugin_name, version).wrapped_dict)


@rest.patch('/plugins/<plugin_name>')
@acl.enforce("data-processing:plugins:patch")
@v.check_exists(api.get_plugin, plugin_name='plugin_name')
@v.validate(v_p.plugin_update_validation_jsonschema(), v_p.check_plugin_update)
def plugins_update(plugin_name, data):
    return u.render(api.update_plugin(plugin_name, data).wrapped_dict)


@rest.post_file('/plugins/<plugin_name>/<version>/convert-config/<name>')
@acl.enforce("data-processing:plugins:convert_config")
@v.check_exists(api.get_plugin, plugin_name='plugin_name', version='version')
@v.validate(None, v_p.check_convert_to_template)
def plugins_convert_to_cluster_template(plugin_name, version, name, data):
    # There is no plugins that supports converting to cluster template
    # The last plugin with support of that is no longer supported
    pass


# Image Registry ops

@rest.get('/images')
@acl.enforce("data-processing:images:get_all")
def images_list():
    tags = u.get_request_args().getlist('tags')
    name = u.get_request_args().get('name', None)
    return u.render(images=[i.dict for i in api.get_images(name, tags)])


@rest.get('/images/<image_id>')
@acl.enforce("data-processing:images:get")
@v.check_exists(api.get_image, id='image_id')
def images_get(image_id):
    return u.render(api.get_registered_image(image_id=image_id).wrapped_dict)


@rest.post('/images/<image_id>')
@acl.enforce("data-processing:images:register")
@v.check_exists(api.get_image, id='image_id')
@v.validate(v_images.image_register_schema, v_images.check_image_register)
def images_set(image_id, data):
    return u.render(api.register_image(image_id, **data).wrapped_dict)


@rest.delete('/images/<image_id>')
@acl.enforce("data-processing:images:unregister")
@v.check_exists(api.get_image, id='image_id')
def images_unset(image_id):
    api.unregister_image(image_id)
    return u.render()


@rest.post('/images/<image_id>/tag')
@acl.enforce("data-processing:images:add_tags")
@v.check_exists(api.get_image, id='image_id')
@v.validate(v_images.image_tags_schema, v_images.check_tags)
def image_tags_add(image_id, data):
    return u.render(api.add_image_tags(image_id, **data).wrapped_dict)


@rest.post('/images/<image_id>/untag')
@acl.enforce("data-processing:images:remove_tags")
@v.check_exists(api.get_image, id='image_id')
@v.validate(v_images.image_tags_schema)
def image_tags_delete(image_id, data):
    return u.render(api.remove_image_tags(image_id, **data).wrapped_dict)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import re

from hacking import core

ALL_LOG_LEVELS = "info|exception|warning|critical|error|debug"

RE_ACCEPTED_LOG_LEVELS = re.compile(
    r"(.)*LOG\.(%(levels)s)\(" % {'levels': ALL_LOG_LEVELS})

# Since _Lx() have been removed, we just need to check _()
RE_TRANSLATED_LOG = re.compile(
    r"(.)*LOG\.(%(levels)s)\(\s*_\(" % {'levels': ALL_LOG_LEVELS})


@core.flake8ext
def no_translate_logs(logical_line, filename):
    """Check for 'LOG.*(_('

    Translators don't provide translations for log messages, and operators
    asked not to translate them.

    * This check assumes that 'LOG' is a logger.
    * Use filename so we can start enforcing this in specific folders instead
      of needing to do so all at once.
    S373
    """

    msg = "S373 Don't translate logs"
    if RE_TRANSLATED_LOG.match(logical_line):
        yield (0, msg)


@core.flake8ext
def accepted_log_levels(logical_line, filename):
    """In Sahara we use only 5 log levels.

    This check is needed because we don't want new contributors to
    use deprecated log levels.
    S374
    """

    # NOTE(Kezar): sahara/tests included because we don't require translations
    # in tests. sahara/db/templates provide separate cli interface so we don't
    # want to translate it.

    ignore_dirs = ["sahara/db/templates",
                   "sahara/tests"]
    for directory in ignore_dirs:
        if directory in filename:
            return
    msg = ("S374 You used deprecated log level. Accepted log levels are "
           "%(levels)s" % {'levels': ALL_LOG_LEVELS})
    if logical_line.startswith("LOG."):
        if not RE_ACCEPTED_LOG_LEVELS.search(logical_line):
            yield(0, msg)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import random

from oslo_config import cfg
from oslo_log import log
from oslo_service import periodic_task
from oslo_service import threadgroup
from oslo_utils import timeutils
import six

from sahara import conductor as c
from sahara import context
from sahara.service.api import v10 as api
from sahara.service import coordinator
from sahara.service.edp import job_manager
from sahara.service.health import verification_base as vb
from sahara.service import trusts
from sahara.utils import cluster as c_u
from sahara.utils import edp
from sahara.utils import proxy as p


LOG = log.getLogger(__name__)

periodic_opts = [
    cfg.BoolOpt('periodic_enable',
                default=True,
                help='Enable periodic tasks.'),
    cfg.IntOpt('periodic_fuzzy_delay',
               default=60,
               help='Range in seconds to randomly delay when starting the'
                    ' periodic task scheduler to reduce stampeding.'
                    ' (Disable by setting to 0).'),
    cfg.IntOpt('periodic_interval_max',
               default=60,
               help='Max interval size between periodic tasks execution in '
                    'seconds.'),
    cfg.IntOpt('min_transient_cluster_active_time',
               default=30,
               help='Minimal "lifetime" in seconds for a transient cluster. '
                    'Cluster is guaranteed to be "alive" within this time '
                    'period.'),
    cfg.IntOpt('cleanup_time_for_incomplete_clusters',
               default=0,
               help='Maximal time (in hours) for clusters allowed to be in '
                    'states other than "Active", "Deleting" or "Error". If a '
                    'cluster is not in "Active", "Deleting" or "Error" state '
                    'and last update of it was longer than '
                    '"cleanup_time_for_incomplete_clusters" hours ago then it '
                    'will be deleted automatically. (0 value means that '
                    'automatic clean up is disabled).'),
    cfg.StrOpt('periodic_coordinator_backend_url',
               help='The backend URL to use for distributed periodic tasks '
                    'coordination.'),
    cfg.IntOpt('periodic_workers_number',
               default=1,
               help='Number of threads to run periodic tasks.'),
]

CONF = cfg.CONF
CONF.register_opts(periodic_opts)

conductor = c.API


def get_time_since_last_update(cluster):
    cluster_updated_at = timeutils.normalize_time(
        timeutils.parse_isotime(cluster.updated_at))
    current_time = timeutils.utcnow()
    spacing = timeutils.delta_seconds(cluster_updated_at,
                                      current_time)
    return spacing


def terminate_cluster(ctx, cluster, description):
    if CONF.use_identity_api_v3 and cluster.trust_id:
        trusts.use_os_admin_auth_token(cluster)
        context.set_current_cluster_id(cluster.id)

        LOG.debug('Terminating {description} cluster '
                  'in "{status}" state'.format(status=cluster.status,
                                               description=description))

        try:
            api.terminate_cluster(cluster.id)
        except Exception as e:
            LOG.warning(
                'Failed to terminate {description} cluster in "{status}" '
                'state: {error}.'.format(error=six.text_type(e),
                                         status=cluster.status,
                                         description=description))

    else:
        if (cluster.status !=
                c_u.CLUSTER_STATUS_AWAITINGTERMINATION):
            conductor.cluster_update(
                ctx, cluster,
                {'status': c_u.CLUSTER_STATUS_AWAITINGTERMINATION})


def set_context(func):
    @functools.wraps(func)
    def handler(self, ctx):
        ctx = context.get_admin_context()
        context.set_ctx(ctx)
        func(self, ctx)
        context.set_ctx(None)
    return handler


def _make_periodic_tasks():
    '''Return the periodic tasks object

    This function creates the periodic tasks class object, it is wrapped in
    this manner to allow easier control of enabling and disabling tasks.

    '''
    zombie_task_spacing = 300 if CONF.use_domain_for_proxy_users else -1
    heartbeat_interval = (CONF.coordinator_heartbeat_interval if
                          CONF.periodic_coordinator_backend_url else -1)

    class SaharaPeriodicTasks(periodic_task.PeriodicTasks):
        hr = coordinator.HashRing(
            CONF.periodic_coordinator_backend_url, 'sahara-periodic-tasks')

        def __init__(self):
            super(SaharaPeriodicTasks, self).__init__(CONF)

        @periodic_task.periodic_task(
            spacing=heartbeat_interval, run_immediately=True)
        @set_context
        def heartbeat(self, ctx):
            self.hr.heartbeat()

        @periodic_task.periodic_task(spacing=45)
        @set_context
        def update_job_statuses(self, ctx):
            LOG.debug('Updating job statuses')
            all_je = conductor.job_execution_get_all(ctx, end_time=None)
            je_to_manage = self.hr.get_subset(all_je)
            for job in je_to_manage:
                job_manager.update_job_status(job.id)

        @periodic_task.periodic_task(spacing=90)
        @set_context
        def terminate_unneeded_transient_clusters(self, ctx):
            LOG.debug('Terminating unneeded transient clusters')
            all_clusters = conductor.cluster_get_all(
                ctx, status=c_u.CLUSTER_STATUS_ACTIVE, is_transient=True)
            clusters_to_manage = self.hr.get_subset(all_clusters)

            for cluster in clusters_to_manage:
                jc = conductor.job_execution_count(ctx,
                                                   end_time=None,
                                                   cluster_id=cluster.id)

                if jc > 0:
                    continue

                spacing = get_time_since_last_update(cluster)
                if spacing < CONF.min_transient_cluster_active_time:
                    continue

                terminate_cluster(ctx, cluster, description='transient')
                # Add event log info cleanup
                context.ctx().current_instance_info = context.InstanceInfo()

        @periodic_task.periodic_task(spacing=zombie_task_spacing)
        @set_context
        def check_for_zombie_proxy_users(self, ctx):
            all_users = p.proxy_domain_users_list()
            users_to_manage = self.hr.get_subset(all_users)
            for user in users_to_manage:
                if user.name.startswith('job_'):
                    je_id = user.name[4:]
                    je = conductor.job_execution_get(ctx, je_id)
                    if je is None or (je.info['status'] in
                                      edp.JOB_STATUSES_TERMINATED):
                        LOG.debug('Found zombie proxy user {username}'.format(
                            username=user.name))
                        p.proxy_user_delete(user_id=user.id)

        @periodic_task.periodic_task(spacing=3600)
        @set_context
        def terminate_incomplete_clusters(self, ctx):
            if CONF.cleanup_time_for_incomplete_clusters <= 0:
                return

            LOG.debug('Terminating old clusters in non-final state')

            # NOTE(alazarev) Retrieving all clusters once in hour for now.
            # Criteria support need to be implemented in sahara db API to
            # have SQL filtering.
            all_clusters = [
                cluster for cluster in conductor.cluster_get_all(ctx) if
                (cluster.status not in [
                    c_u.CLUSTER_STATUS_ACTIVE, c_u.CLUSTER_STATUS_ERROR,
                    c_u.CLUSTER_STATUS_DELETING])
                ]
            clusters_to_manage = self.hr.get_subset(all_clusters)

            for cluster in clusters_to_manage:

                spacing = get_time_since_last_update(cluster)
                if spacing < CONF.cleanup_time_for_incomplete_clusters * 3600:
                    continue

                terminate_cluster(ctx, cluster, description='incomplete')
                # Add event log info cleanup
                context.ctx().current_instance_info = context.InstanceInfo()

        @periodic_task.periodic_task(
            spacing=vb.get_verification_periodic_interval())
        @set_context
        def run_verifications(self, ctx):
            LOG.debug("Executing health checks for the clusters")
            start_dict = {'verification': {'status': 'START'}}
            all_clusters = conductor.cluster_get_all(
                ctx, status=c_u.CLUSTER_STATUS_ACTIVE)
            clusters_to_manage = self.hr.get_subset(all_clusters)

            for cluster in clusters_to_manage:
                try:
                    vb.validate_verification_start(cluster)
                    api.update_cluster(cluster.id, start_dict)
                except vb.CannotVerifyError:
                    LOG.debug("Skipping running verification "
                              "on the cluster %s", cluster.name)

    return SaharaPeriodicTasks()


def setup():
    if CONF.periodic_enable:
        if CONF.periodic_fuzzy_delay:
            initial_delay = random.randint(0, CONF.periodic_fuzzy_delay)
            LOG.debug("Starting periodic tasks with initial delay {seconds} "
                      "seconds".format(seconds=initial_delay))
        else:
            initial_delay = None

        tg = threadgroup.ThreadGroup()
        workers_number = (CONF.periodic_workers_number
                          if CONF.periodic_coordinator_backend_url else 1)

        for t in range(workers_number):
            pt = _make_periodic_tasks()
            tg.add_dynamic_timer(
                pt.run_periodic_tasks,
                initial_delay=initial_delay,
                periodic_interval_max=CONF.periodic_interval_max,
                context=None)



# Copyright 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""032_add_domain_name

Revision ID: 032
Revises: 031
Create Date: 2016-07-21 13:33:33.674853

"""

# revision identifiers, used by Alembic.
revision = '032'
down_revision = '031'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('cluster_templates', sa.Column(
        'domain_name', sa.String(length=255), nullable=True))
    op.add_column('clusters', sa.Column(
        'domain_name', sa.String(length=255), nullable=True))
    op.add_column('instances', sa.Column(
        'dns_hostname', sa.String(length=255), nullable=True))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from sahara import conductor as c
from sahara import context
from sahara import exceptions as e
from sahara.i18n import _
from sahara.service.edp import job_utils
from sahara.service.edp.oozie import engine as oozie_engine
from sahara.service.edp.spark import engine as spark_engine
from sahara.service.edp.storm import engine as storm_engine
from sahara.utils import cluster as c_u
from sahara.utils import edp
from sahara.utils import proxy as p


LOG = log.getLogger(__name__)

CONF = cfg.CONF

conductor = c.API

ENGINES = [oozie_engine.OozieJobEngine,
           spark_engine.SparkJobEngine,
           storm_engine.StormJobEngine,
           storm_engine.StormPyleusJobEngine]


def _get_job_type(job_execution):
    return conductor.job_get(context.ctx(), job_execution.job_id).type


def get_job_engine(cluster, job_execution):
    return job_utils.get_plugin(cluster).get_edp_engine(cluster,
                                                        _get_job_type(
                                                            job_execution))


def _write_job_status(job_execution, job_info):
    update = {"info": job_info}
    if job_info['status'] in edp.JOB_STATUSES_TERMINATED:
        update['end_time'] = datetime.datetime.now()
        job_configs = p.delete_proxy_user_for_job_execution(job_execution)
        if job_configs:
            update['job_configs'] = job_configs
    return conductor.job_execution_update(context.ctx(),
                                          job_execution,
                                          update)


def _update_job_status(engine, job_execution):
    job_info = engine.get_job_status(job_execution)
    if job_info is not None:
        job_execution = _write_job_status(job_execution, job_info)
    return job_execution


def _update_job_execution_extra(cluster, job_execution):
    # tmckay-fp we can make this slightly more efficient in
    # the use_namespaces case by asking the engine if it knows
    # the submission machine, and checking if that machine has
    # a floating ip.
    if (CONF.use_namespaces or CONF.proxy_command):
        info = cluster.node_groups[0].instances[0].remote().get_neutron_info()
        extra = job_execution.extra.copy()
        extra['neutron'] = info

        job_execution = conductor.job_execution_update(
            context.ctx(), job_execution.id, {'extra': extra})
    return job_execution


def _run_job(job_execution_id):
    ctx = context.ctx()
    job_execution = conductor.job_execution_get(ctx, job_execution_id)
    cluster = conductor.cluster_get(ctx, job_execution.cluster_id)
    if cluster is None or cluster.status != c_u.CLUSTER_STATUS_ACTIVE:
        LOG.info("Can not run this job on a non-existant cluster or a "
                 "inactive cluster.")
        return

    eng = get_job_engine(cluster, job_execution)
    if eng is None:
        raise e.EDPError(_("Cluster does not support job type %s")
                         % _get_job_type(job_execution))
    job_execution = _update_job_execution_extra(cluster, job_execution)

    # Job id is a string
    # Status is a string
    # Extra is a dictionary to add to extra in the job_execution
    if job_execution.job_configs.job_execution_info.get('job_execution_type'
                                                        ) == 'scheduled':
        jid, status, extra = eng.run_scheduled_job(job_execution)
    else:
        jid, status, extra = eng.run_job(job_execution)

    # Set the job id and the start time
    # Optionally, update the status and the 'extra' field
    update_dict = {'engine_job_id': jid,
                   'start_time': datetime.datetime.now()}
    if status:
        update_dict['info'] = {'status': status}
    if extra:
        curr_extra = job_execution.extra.copy()
        if 'neutron' in curr_extra:
            curr_extra['neutron'] = curr_extra['neutron'].copy()
        curr_extra.update(extra)
        update_dict['extra'] = curr_extra

    job_execution = conductor.job_execution_update(
        ctx, job_execution, update_dict)


def run_job(job_execution_id):
    try:
        _run_job(job_execution_id)
    except Exception as ex:
        LOG.exception("Can't run job execution (reason: {reason})".format(
                      reason=ex))

        job_execution = conductor.job_execution_get(
            context.ctx(), job_execution_id)

        if job_execution.engine_job_id is not None:
            cancel_job(job_execution_id)

        conductor.job_execution_update(
            context.ctx(), job_execution_id,
            {'info': {'status': edp.JOB_STATUS_FAILED},
             'start_time': datetime.datetime.now(),
             'end_time': datetime.datetime.now()})


def cancel_job(job_execution_id):
    ctx = context.ctx()
    job_execution = conductor.job_execution_get(ctx, job_execution_id)
    if job_execution.info['status'] in edp.JOB_STATUSES_TERMINATED:
        LOG.info("Job execution is already finished and shouldn't be canceled")
        return job_execution
    cluster = conductor.cluster_get(ctx, job_execution.cluster_id)
    if cluster is None:
        LOG.info("Can not cancel this job on a non-existant cluster.")
        return job_execution
    engine = get_job_engine(cluster, job_execution)
    if engine is not None:
        job_execution = conductor.job_execution_update(
            ctx, job_execution_id,
            {'info': {'status': edp.JOB_STATUS_TOBEKILLED}})

        timeout = CONF.job_canceling_timeout
        s_time = timeutils.utcnow()
        while timeutils.delta_seconds(s_time, timeutils.utcnow()) < timeout:
            if job_execution.info['status'] not in edp.JOB_STATUSES_TERMINATED:
                try:
                    job_info = engine.cancel_job(job_execution)
                except Exception as ex:
                    job_info = None
                    LOG.warning("Error during cancel of job execution: "
                                "{error}".format(error=ex))
                if job_info is not None:
                    job_execution = _write_job_status(job_execution, job_info)
                    LOG.info("Job execution was canceled successfully")
                    return job_execution
                context.sleep(3)
                job_execution = conductor.job_execution_get(
                    ctx, job_execution_id)
                if not job_execution:
                    LOG.info("Job execution was deleted. "
                             "Canceling current operation.")
                    return job_execution
            else:
                LOG.info("Job execution status: {status}").format(
                    status=job_execution.info['status'])
                return job_execution
        else:
            raise e.CancelingFailed(_('Job execution %s was not canceled')
                                    % job_execution.id)


def get_job_status(job_execution_id):
    ctx = context.ctx()
    job_execution = conductor.job_execution_get(ctx, job_execution_id)
    cluster = conductor.cluster_get(ctx, job_execution.cluster_id)
    if (cluster is not None and
            cluster.status == c_u.CLUSTER_STATUS_ACTIVE):
        engine = get_job_engine(cluster, job_execution)
        if engine is not None:
            job_execution = _update_job_status(engine,
                                               job_execution)
    return job_execution


def update_job_status(job_execution_id):
    try:
        get_job_status(job_execution_id)
    except Exception as e:
        LOG.exception("Error during update job execution {job}: {error}"
                      .format(job=job_execution_id, error=e))


def update_job_statuses(cluster_id=None):
    ctx = context.ctx()
    kwargs = {'end_time': None}
    if cluster_id:
        kwargs.update({'cluster_id': cluster_id})
    for je in conductor.job_execution_get_all(ctx, **kwargs):
        update_job_status(je.id)


def get_job_config_hints(job_type):
    for eng in ENGINES:
        if job_type in eng.get_supported_job_types():
            return eng.get_possible_job_config(job_type)


def suspend_job(job_execution_id):
    ctx = context.ctx()
    job_execution = conductor.job_execution_get(ctx, job_execution_id)
    if job_execution.info['status'] not in edp.JOB_STATUSES_SUSPENDIBLE:
        raise e.SuspendingFailed(_("Suspending operation can not be performed"
                                 " on status: {status}")).format(
                                     status=job_execution.info['status'])
    cluster = conductor.cluster_get(ctx, job_execution.cluster_id)
    engine = get_job_engine(cluster, job_execution)
    job_execution = conductor.job_execution_update(
        ctx, job_execution_id, {
            'info': {'status': edp.JOB_STATUS_TOBESUSPENDED}})
    try:
        job_info = engine.suspend_job(job_execution)
    except Exception as ex:
        job_info = None
        conductor.job_execution_update(
            ctx, job_execution_id, {'info': {
                'status': edp.JOB_STATUS_SUSPEND_FAILED}})
        raise e.SuspendingFailed(_("Error during suspending of job execution: "
                                   "{error}")).format(error=ex)
    if job_info is not None:
        job_execution = _write_job_status(job_execution, job_info)
        LOG.info("Job execution was suspended successfully")
        return job_execution

    conductor.job_execution_update(
        ctx, job_execution_id, {'info': {
            'status': edp.JOB_STATUS_SUSPEND_FAILED}})
    raise e.SuspendingFailed(_("Failed to suspend job execution "
                               "{jid}")).format(jid=job_execution_id)



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.api import acl
from sahara.service.api.v2 import job_binaries as api
from sahara.service import validation as v
from sahara.service.validations.edp import job_binary as v_j_b
from sahara.service.validations.edp import job_binary_schema as v_j_b_schema
import sahara.utils.api as u


rest = u.RestV2('job-binaries', __name__)


@rest.post('/job-binaries')
@acl.enforce("data-processing:job-binary:create")
@v.validate(v_j_b_schema.JOB_BINARY_SCHEMA, v_j_b.check_job_binary)
@v.validate_request_params([])
def job_binary_create(data):
    result = api.create_job_binary(data).to_wrapped_dict()
    u._replace_tenant_id_project_id(result['job_binary'])
    return u.render(result)


@rest.get('/job-binaries')
@acl.enforce("data-processing:job-binary:list")
@v.check_exists(api.get_job_binary, 'marker')
@v.validate(None, v.validate_pagination_limit,
            v.validate_sorting_job_binaries)
@v.validate_request_params(['name'])
def job_binary_list():
    result = api.get_job_binaries(**u.get_request_args().to_dict())
    for jb in result:
        u._replace_tenant_id_project_id(jb)
    return u.render(res=result, name='binaries')


@rest.get('/job-binaries/<job_binary_id>')
@acl.enforce("data-processing:job-binary:get")
@v.check_exists(api.get_job_binary, 'job_binary_id')
@v.validate_request_params([])
def job_binary_get(job_binary_id):
    result = api.get_job_binary(job_binary_id).to_wrapped_dict()
    u._replace_tenant_id_project_id(result['job_binary'])
    return u.render(result)


@rest.delete('/job-binaries/<job_binary_id>')
@acl.enforce("data-processing:job-binary:delete")
@v.check_exists(api.get_job_binary, id='job_binary_id')
@v.validate_request_params([])
def job_binary_delete(job_binary_id):
    api.delete_job_binary(job_binary_id)
    return u.render()


@rest.get('/job-binaries/<job_binary_id>/data')
@acl.enforce("data-processing:job-binary:get-data")
@v.check_exists(api.get_job_binary, 'job_binary_id')
@v.validate_request_params([])
def job_binary_data(job_binary_id):
    data = api.get_job_binary_data(job_binary_id)
    if type(data) == dict:
        data = u.render(data)
    return data


@rest.patch('/job-binaries/<job_binary_id>')
@acl.enforce("data-processing:job-binary:update")
@v.validate(v_j_b_schema.JOB_BINARY_UPDATE_SCHEMA, v_j_b.check_job_binary)
@v.validate_request_params([])
def job_binary_update(job_binary_id, data):
    result = api.update_job_binary(job_binary_id, data).to_wrapped_dict()
    u._replace_tenant_id_project_id(result['job_binary'])
    return u.render(result)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from oslo_config import cfg

from sahara import conductor as c
from sahara import context
import sahara.exceptions as ex
from sahara.i18n import _
import sahara.service.edp.data_sources.manager as ds_manager
import sahara.service.validations.edp.base as b


CONF = cfg.CONF


def check_data_source_create(data, **kwargs):
    b.check_data_source_unique_name(data['name'])
    _check_data_source(data)


def _check_datasource_placeholder(url):
    if url is None:
        return
    total_length = 0
    substrings = re.findall(r"%RANDSTR\(([\-]?\d+)\)%", url)
    for length in map(int, substrings):
        if length <= 0:
            raise ex.InvalidDataException(_("Requested RANDSTR length"
                                            " must be positive."))
        total_length += length

    if total_length > 1024:
        raise ex.InvalidDataException(_("Requested RANDSTR length is"
                                        " too long, please choose a "
                                        "value less than 1024."))


def _check_data_source(data):
    _check_datasource_placeholder(data["url"])
    if data["type"] in CONF.data_source_types:
        ds_manager.DATA_SOURCES.get_data_source(data["type"]).validate(data)


def check_data_source_update(data, data_source_id):
    ctx = context.ctx()
    jobs = c.API.job_execution_get_all(ctx)
    pending_jobs = [job for job in jobs if job.info["status"] == "PENDING"]
    for job in pending_jobs:
        if data_source_id in job.data_source_urls:
            raise ex.UpdateFailedException(
                _("DataSource is used in a "
                  "PENDING Job and can not be updated."))

    ds = c.API.data_source_get(ctx, data_source_id)
    if 'name' in data and data['name'] != ds.name:
        b.check_data_source_unique_name(data['name'])

    check_data = {'type': data.get('type', None) or ds.type,
                  'url': data.get('url', None) or ds.url,
                  'credentials': data.get(
                      'credentials', None) or ds.credentials}
    _check_data_source(check_data)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import string

from oslo_utils import uuidutils
import six

from sahara.i18n import _


class SaharaException(Exception):
    """Base Exception for the project

    To correctly use this class, inherit from it and define
    a 'message' and 'code' properties.
    """
    code = "UNKNOWN_EXCEPTION"
    message = _("An unknown exception occurred")

    def __str__(self):
        return self.message

    def __init__(self, message=None, code=None, inject_error_id=True):
        self.uuid = uuidutils.generate_uuid()

        if code:
            self.code = code
        if message:
            self.message = message

        if inject_error_id:
            # Add Error UUID to the message if required
            self.message = (_('%(message)s\nError ID: %(id)s')
                            % {'message': self.message, 'id': self.uuid})

        super(SaharaException, self).__init__(
            '%s: %s' % (self.code, self.message))


class NotFoundException(SaharaException):
    code = "NOT_FOUND"
    message_template = _("Object '%s' is not found")

    # It could be a various property of object which was not found
    def __init__(self, value, message_template=None):
        self.value = value
        if message_template:
            formatted_message = message_template % value
        else:
            formatted_message = self.message_template % value

        super(NotFoundException, self).__init__(formatted_message)


class NoUniqueMatchException(SaharaException):
    code = "NO_UNIQUE_MATCH"
    message_template = _(
        "Response {response} is not unique for this query {query}.")

    def __init__(self, response, query, message_template=None):
        template = message_template or self.message_template
        formatted_message = template.format(response=response, query=query)
        super(NoUniqueMatchException, self).__init__(formatted_message)


class NameAlreadyExistsException(SaharaException):
    code = "NAME_ALREADY_EXISTS"
    message = _("Name already exists")


class InvalidCredentials(SaharaException):
    message = _("Invalid credentials")
    code = "INVALID_CREDENTIALS"


class InvalidReferenceException(SaharaException):
    code = "INVALID_REFERENCE"
    message = _("Invalid object reference")


class RemoteCommandException(SaharaException):
    code = "REMOTE_COMMAND_FAILED"
    message_template = _("Error during command execution: \"%s\"")

    def __init__(self, cmd, ret_code=None, stdout=None,
                 stderr=None):
        self.cmd = cmd
        self.ret_code = ret_code
        self.stdout = stdout
        self.stderr = stderr

        formatted_message = self.message_template % cmd

        def to_printable(s):
            return "".join(filter(lambda x: x in string.printable, s))

        if ret_code:
            formatted_message = '%s\nReturn code: %s' % (
                formatted_message, six.text_type(ret_code))

        if stderr:
            formatted_message = '%s\nSTDERR:\n%s' % (
                formatted_message, to_printable(stderr))

        if stdout:
            formatted_message = '%s\nSTDOUT:\n%s' % (
                formatted_message, to_printable(stdout))

        super(RemoteCommandException, self).__init__(formatted_message)


class InvalidDataException(SaharaException):
    """General exception to use for invalid data

    A more useful message should be passed to __init__ which
    tells the user more about why the data is invalid.
    """
    code = "INVALID_DATA"
    message = _("Data is invalid")


class BadJobBinaryInternalException(SaharaException):
    code = "BAD_JOB_BINARY"
    message = _("Job binary internal data must be a string of length "
                "greater than zero")


class BadJobBinaryException(SaharaException):
    code = "BAD_JOB_BINARY"
    message = _("To work with JobBinary located in internal swift add 'user'"
                " and 'password' to extra")


class DBDuplicateEntry(SaharaException):
    code = "DB_DUPLICATE_ENTRY"
    message = _("Database object already exists")


class CreationFailed(SaharaException):
    message = _("Object was not created")
    code = "CREATION_FAILED"


class CancelingFailed(SaharaException):
    message = _("Operation was not canceled")
    code = "CANCELING_FAILED"


class SuspendingFailed(SaharaException):
    message = _("Operation was not suspended")
    code = "SUSPENDING_FAILED"


class InvalidJobStatus(SaharaException):
    message = _("Invalid Job Status")
    code = "INVALID_JOB_STATUS"


class DeletionFailed(SaharaException):
    code = "DELETION_FAILED"
    message = _("Object was not deleted")


class MissingFloatingNetworkException(SaharaException):
    code = "MISSING_FLOATING_NETWORK"
    message_template = _("Node Group %s is missing 'floating_ip_pool' "
                         "field")

    def __init__(self, ng_name):
        formatted_message = self.message_template % ng_name

        super(MissingFloatingNetworkException, self).__init__(
            formatted_message)


class SwiftClientException(SaharaException):
    '''General wrapper object for swift client exceptions

    This exception is intended for wrapping the message from a
    swiftclient.ClientException in a SaharaException. The ClientException
    should be caught and an instance of SwiftClientException raised instead.
    '''

    code = "SWIFT_CLIENT_EXCEPTION"
    message = _("An error has occurred while performing a request to Swift")


class S3ClientException(SaharaException):
    '''General wrapper object for boto exceptions

    Intended to replace any errors raised by the botocore client.
    '''

    code = "S3_CLIENT_EXCEPTION"
    message = _("An error has occurred while performing a request to S3")


class DataTooBigException(SaharaException):
    code = "DATA_TOO_BIG"
    message_template = _("Size of data (%(size)s) is greater than maximum "
                         "(%(maximum)s)")

    def __init__(self, size, maximum, message_template=None):
        if message_template:
            self.message_template = message_template

        formatted_message = self.message_template % (
            {'size': size, 'maximum': maximum})

        super(DataTooBigException, self).__init__(formatted_message)


class ThreadException(SaharaException):
    code = "THREAD_EXCEPTION"
    message_template = _("An error occurred in thread '%(thread)s': %(e)s"
                         "\n%(stacktrace)s")

    def __init__(self, thread_description, e, stacktrace):
        formatted_message = self.message_template % {
            'thread': thread_description,
            'e': six.text_type(e),
            'stacktrace': stacktrace}

        super(ThreadException, self).__init__(formatted_message)


class SubprocessException(SaharaException):
    code = "SUBPROCESS_EXCEPTION"
    message = _("Subprocess execution has failed")


class NotImplementedException(SaharaException):
    code = "NOT_IMPLEMENTED"
    message_template = _("Feature '%s' is not implemented")

    def __init__(self, feature, message_template=None):
        if message_template:
            self.message_template = message_template

        formatted_message = self.message_template % feature

        super(NotImplementedException, self).__init__(formatted_message)


class HeatStackException(SaharaException):
    code = "HEAT_STACK_EXCEPTION"
    message_template = _("Heat stack failed with status %s")

    def __init__(self, heat_stack_status=None, message=None):
        if message:
            formatted_message = message
        elif heat_stack_status:
            formatted_message = self.message_template % heat_stack_status
        else:
            formatted_message = _("Heat stack failed")
        super(HeatStackException, self).__init__(formatted_message)


class ConfigurationError(SaharaException):
    code = "CONFIGURATION_ERROR"
    message = _("The configuration has failed")


class IncorrectStateError(SaharaException):
    message = _("The object is in an incorrect state")
    code = "INCORRECT_STATE_ERROR"


class FrozenClassError(SaharaException):
    code = "FROZEN_CLASS_ERROR"
    message_template = _("Class %s is immutable!")

    def __init__(self, instance):
        formatted_message = self.message_template % type(instance).__name__

        super(FrozenClassError, self).__init__(formatted_message)


class SystemError(SaharaException):
    code = "SYSTEM_ERROR"
    message = _("System error has occurred")


class EDPError(SaharaException):
    code = "EDP_ERROR"
    message = _("Failed to complete EDP operation")


class OozieException(SaharaException):
    code = "OOZIE_EXCEPTION"
    message = _("Failed to perform Oozie request")


class TimeoutException(SaharaException):
    code = "TIMEOUT"
    message_template = _("'%(operation)s' timed out after %(timeout)i "
                         "second(s)")

    def __init__(self, timeout, op_name=None, timeout_name=None):
        if op_name:
            op_name = _("Operation with name '%s'") % op_name
        else:
            op_name = _("Operation")
        formatted_message = self.message_template % {
            'operation': op_name, 'timeout': timeout}

        if timeout_name:
            desc = _("%(message)s and following timeout was violated: "
                     "%(timeout_name)s")
            formatted_message = desc % {
                'message': formatted_message, 'timeout_name': timeout_name}

        super(TimeoutException, self).__init__(formatted_message)


class DeprecatedException(SaharaException):
    code = "DEPRECATED"
    message = _("The version you are trying to use is deprecated")


class Forbidden(SaharaException):
    code = "FORBIDDEN"
    message = _("You are not authorized to complete this action")


class ImageNotRegistered(SaharaException):
    code = "IMAGE_NOT_REGISTERED"
    message_template = _("Image %s is not registered in Sahara")

    def __init__(self, image):
        formatted_message = self.message_template % image
        super(ImageNotRegistered, self).__init__(formatted_message)


class MalformedRequestBody(SaharaException):
    code = "MALFORMED_REQUEST_BODY"
    message_template = _("Malformed message body: %(reason)s")

    def __init__(self, reason):
        formatted_message = self.message_template % {"reason": reason}
        super(MalformedRequestBody, self).__init__(formatted_message)


class QuotaException(SaharaException):
    code = "QUOTA_ERROR"
    message_template = _("Quota exceeded for %(resource)s: "
                         "Requested %(requested)s, "
                         "but available %(available)s")

    def __init__(self, resource, requested, available):
        formatted_message = self.message_template % {
            'resource': resource,
            'requested': requested,
            'available': available}

        super(QuotaException, self).__init__(formatted_message)


class UpdateFailedException(SaharaException):
    code = "UPDATE_FAILED"
    message_template = _("Object '%s' could not be updated")
    # Object was unable to be updated

    def __init__(self, value, message_template=None):
        if message_template:
            self.message_template = message_template

        formatted_message = self.message_template % value

        super(UpdateFailedException, self).__init__(formatted_message)


class MaxRetriesExceeded(SaharaException):
    code = "MAX_RETRIES_EXCEEDED"
    message_template = _("Operation %(operation)s wasn't executed correctly "
                         "after %(attempts)d attempts")

    def __init__(self, attempts, operation):
        formatted_message = self.message_template % {'operation': operation,
                                                     'attempts': attempts}

        super(MaxRetriesExceeded, self).__init__(formatted_message)


class InvalidJobExecutionInfoException(SaharaException):
    message = _("Job execution information is invalid")

    def __init__(self, message=None):
        if message:
            self.message = message
        self.code = "INVALID_JOB_EXECUTION_INFO"
        super(InvalidJobExecutionInfoException, self).__init__()



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import functools
import threading

from eventlet import timeout as e_timeout
from oslo_config import cfg
from oslo_log import log as logging
import six

from sahara import conductor
from sahara import context
from sahara import exceptions as ex
from sahara.i18n import _
from sahara.plugins import base as plugin_base
from sahara.service.health import common
from sahara.utils import cluster as cluster_utils
from sahara.utils.notification import sender

cond = conductor.API
CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class BaseHealthError(ex.SaharaException):
    message_template = _("Cluster health is %(status)s. Reason: %(reason)s")
    code = 'HEALTH_ERROR'
    status = 'UNKNOWN'

    def __init__(self, reason):
        message = self.message_template % {
            'status': self.status, 'reason': reason}
        # Ignoring Error id because it's not really needed
        super(BaseHealthError, self).__init__(message, inject_error_id=False)


class RedHealthError(BaseHealthError):
    """Exception to indicate red state of the health check."""
    code = "RED_STATE"
    status = common.HEALTH_STATUS_RED


class YellowHealthError(BaseHealthError):
    """Exception to indicate yellow state of health check."""
    code = "YELLOW_STATE"
    status = common.HEALTH_STATUS_YELLOW


@six.add_metaclass(abc.ABCMeta)
class BasicHealthCheck(object):
    def __init__(self, cluster):
        """Initialize a health check for the specified cluster."""
        self.cluster = cluster
        self.health_check_id = None

    @abc.abstractmethod
    def get_health_check_name(self):
        """Return name of the health check."""
        pass

    @abc.abstractmethod
    def is_available(self):
        """Validate availability of the health check for the specified cluster

        Return True when the health check is available for the specified
        cluster and False when it's not.
        """
        pass

    @abc.abstractmethod
    def check_health(self):
        """Check health of the specified cluster

        Returns description if health check executed successfully. Raises
        YellowStateError to indicate Yellow result of the health check,
        and RedStateError to indicate Red result of the health check.
        """
        pass

    def _indicate_start(self):
        vid = self.cluster.verification.id
        self.health_check_id = cond.cluster_health_check_add(
            context.ctx(), vid, {'status': common.HEALTH_STATUS_CHECKING,
                                 'name': self.get_health_check_name()}).id
        self.health_check = cond.cluster_health_check_get(
            context.ctx(), self.health_check_id)
        sender.health_notify(self.cluster, self.health_check)

    def _write_result(self, status, description):
        cond.cluster_health_check_update(
            context.ctx(), self.health_check_id,
            {'status': status, 'description': description})
        self.health_check = cond.cluster_health_check_get(
            context.ctx(), self.health_check_id)
        sender.health_notify(self.cluster, self.health_check)

    def execute(self):
        timeout = CONF.cluster_verifications.verification_timeout
        try:
            with e_timeout.Timeout(timeout, ex.TimeoutException(timeout)):
                if not self.is_available():
                    return
                self._indicate_start()
                try:
                    result = self.check_health()
                    status = common.HEALTH_STATUS_GREEN
                except Exception as exc:
                    result = six.text_type(exc)
                    if isinstance(exc, BaseHealthError):
                        status = exc.status
                    else:
                        status = common.HEALTH_STATUS_RED
        except ex.TimeoutException:
            result = _("Health check timed out")
            status = common.HEALTH_STATUS_YELLOW
        self._write_result(status, result)


class AllInstancesAccessible(BasicHealthCheck):
    def __init__(self, cluster, provider):
        self.provider = provider
        super(AllInstancesAccessible, self).__init__(cluster)

    def is_available(self):
        # always available : )
        return True

    def get_health_check_name(self):
        return "Check of instances accessibility"

    def check_health(self):
        inst_ips_or_names = self.provider.get_accessibility_data()
        if inst_ips_or_names:
            insts = ', '.join(inst_ips_or_names)
            LOG.exception(
                "Instances (%s) are not available in the cluster", insts)
            raise RedHealthError(
                _("Instances (%s) are not available in the cluster.") % insts)
        return _("All instances are available")


class ResolvConfIsUnchanged(BasicHealthCheck):
    def __init__(self, cluster, provider):
        self.provider = provider
        super(ResolvConfIsUnchanged, self).__init__(cluster)

    def is_available(self):
        return True

    def get_health_check_name(self):
        return "Check of '/etc/resolv.conf' files"

    def check_health(self):
        bad_inst, bad_res_conf = self.provider.get_resolv_conf_data()
        bad_inst_msg = ''
        res_conf_msg = ''
        if bad_inst:
            insts = ', '.join(bad_inst)
            bad_inst_msg = _("Couldn't read '/etc/resolv.conf' "
                             "on instances: {}.").format(insts)
        if bad_res_conf:
            insts = ', '.join(bad_res_conf)
            ns = ', '.join(CONF.nameservers)
            res_conf_msg = _(
                "Instances ({}) have incorrect '/etc/resolv.conf' "
                "file, expected nameservers: {}.").format(insts, ns)
        if bad_inst_msg or res_conf_msg:
            LOG.exception("{} {}".format(res_conf_msg, bad_inst_msg))
            raise RedHealthError(_("{} {}").format(res_conf_msg, bad_inst_msg))
        return _("All instances have correct '/etc/resolv.conf' file")


class AlertsProvider(object):
    def __init__(self, cluster):
        self._data = None
        self._cluster = cluster
        self._instances = None
        self.get_alerts_data()

    def _instance_get_data(self, instance, lock):
        try:
            with instance.remote() as r:
                data = self._get_resolv_conf(r)
        except Exception:
            data = None
            LOG.exception("Couldn't read '/etc/resolv.conf'")
        with lock:
            self._data[instance.get_ip_or_dns_name()] = data

    def get_accessibility_data(self):
        bad_instances = []
        for el in self._data:
            if self._data[el] is None:
                bad_instances.append(el)
        return bad_instances

    def get_resolv_conf_data(self):
        bad_instances = []
        bad_resolv_conf = []
        for inst_ip_or_name, data in self._data.iteritems():
            if data is None:
                bad_instances.append(inst_ip_or_name)
                continue
            for nameserver in CONF.nameservers:
                if nameserver not in data:
                    bad_resolv_conf.append(inst_ip_or_name)
                    break
        return bad_instances, bad_resolv_conf

    @staticmethod
    def _get_resolv_conf(inst_r):
        # returns None if error occurred while reading resolv.conf
        # otherwise returns content of this file
        code, resolv_conf = inst_r.execute_command(
            "cat /etc/resolv.conf", raise_when_error=False)
        if code != 0:
            return None
        return resolv_conf

    def get_alerts_data(self, check_type=None):
        if check_type and self._data is not None:
            # return cached data
            return self._data.get(check_type, [])
        self._data = {}
        self._instances = cluster_utils.get_instances(self._cluster)
        lock = threading.Lock()
        with context.ThreadGroup() as tg:
            for ins in self._instances:
                tg.spawn('Get health check data of instance %s' % ins.id,
                         self._instance_get_data, ins, lock)
        return self._data


def get_basic(cluster):
    provider = AlertsProvider(cluster)
    basic = [functools.partial(AllInstancesAccessible, provider=provider)]
    if cluster.use_designate_feature():
        basic.append(functools.partial(
            ResolvConfIsUnchanged, provider=provider))
    return basic


def get_health_checks(cluster):
    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)
    plugin_specific = plugin.get_health_checks(cluster)
    plugin_specific.extend(get_basic(cluster))
    return plugin_specific



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from heatclient import exc as heat_exc
from oslo_config import cfg
from oslo_log import log as logging

from sahara import conductor as c
from sahara import context
from sahara import exceptions as ex
from sahara.i18n import _
from sahara.service import engine as e
from sahara.service.heat import commons as heat_common
from sahara.service.heat import templates as ht
from sahara.service import volumes
from sahara.utils import cluster as c_u
from sahara.utils import cluster_progress_ops as cpo
from sahara.utils.openstack import heat

conductor = c.API
CONF = cfg.CONF
LOG = logging.getLogger(__name__)

CREATE_STAGES = [c_u.CLUSTER_STATUS_SPAWNING, c_u.CLUSTER_STATUS_WAITING,
                 c_u.CLUSTER_STATUS_PREPARING]
SCALE_STAGES = [c_u.CLUSTER_STATUS_SCALING_SPAWNING,
                c_u.CLUSTER_STATUS_SCALING_WAITING,
                c_u.CLUSTER_STATUS_SCALING_PREPARING]
ROLLBACK_STAGES = [c_u.CLUSTER_STATUS_ROLLBACK_SPAWNING,
                   c_u.CLUSTER_STATUS_ROLLBACK_WAITING,
                   c_u.CLUSTER_STATUS_ROLLBACK__PREPARING]

heat_engine_opts = [
    cfg.ListOpt('heat_stack_tags', default=['data-processing-cluster'],
                help="List of tags to be used during operating with stack.")
]

CONF.register_opts(heat_engine_opts)


class HeatEngine(e.Engine):
    def get_type_and_version(self):
        return heat_common.HEAT_ENGINE_VERSION

    def create_cluster(self, cluster):
        self._update_rollback_strategy(cluster, shutdown=True)

        target_count = self._get_ng_counts(cluster)
        self._nullify_ng_counts(cluster)
        cluster = self._generate_heat_stack_name(cluster)
        self._launch_instances(cluster, target_count, CREATE_STAGES)

        self._update_rollback_strategy(cluster)

    @staticmethod
    def _generate_heat_stack_name(cluster):
        cluster = conductor.cluster_get(context.ctx(), cluster)
        hsn = cluster.name + cluster.id[:8]
        extra = cluster.extra.to_dict() if cluster.extra else {}
        extra['heat_stack_name'] = hsn
        conductor.cluster_update(context.ctx(), cluster, {'extra': extra})
        return conductor.cluster_get(context.ctx(), cluster)

    def _get_ng_counts(self, cluster):
        count = {}
        for node_group in cluster.node_groups:
            count[node_group.id] = node_group.count
        return count

    def _nullify_ng_counts(self, cluster):
        ctx = context.ctx()

        for node_group in cluster.node_groups:
            conductor.node_group_update(ctx, node_group, {"count": 0})

    def scale_cluster(self, cluster, target_count, instances_to_delete=None):
        ctx = context.ctx()

        rollback_count = self._get_ng_counts(cluster)

        self._update_rollback_strategy(cluster, rollback_count=rollback_count,
                                       target_count=target_count)

        inst_ids = self._launch_instances(
            cluster, target_count, SCALE_STAGES,
            update_stack=True, disable_rollback=False,
            instances_to_delete=instances_to_delete)

        cluster = conductor.cluster_get(ctx, cluster)
        c_u.clean_cluster_from_empty_ng(cluster)

        self._update_rollback_strategy(cluster)

        return inst_ids

    def rollback_cluster(self, cluster, reason):
        rollback_info = cluster.rollback_info or {}
        self._update_rollback_strategy(cluster)

        if rollback_info.get('shutdown', False):
            self._rollback_cluster_creation(cluster, reason)
            LOG.warning("Cluster creation rollback "
                        "(reason: {reason})".format(reason=reason))

            return False

        rollback_count = rollback_info.get('rollback_count', {}).copy()
        target_count = rollback_info.get('target_count', {}).copy()
        if rollback_count or target_count:
            self._rollback_cluster_scaling(
                cluster, rollback_count, target_count, reason)
            LOG.warning("Cluster scaling rollback "
                        "(reason: {reason})".format(reason=reason))

            return True

        return False

    def _update_rollback_strategy(self, cluster, shutdown=False,
                                  rollback_count=None, target_count=None):
        rollback_info = {}

        if shutdown:
            rollback_info['shutdown'] = shutdown

        if rollback_count:
            rollback_info['rollback_count'] = rollback_count

        if target_count:
            rollback_info['target_count'] = target_count

        cluster = conductor.cluster_update(
            context.ctx(), cluster, {'rollback_info': rollback_info})
        return cluster

    def _populate_cluster(self, cluster, stack):
        ctx = context.ctx()
        old_ids = [i.instance_id for i in c_u.get_instances(cluster)]
        new_ids = []

        for node_group in cluster.node_groups:
            instances = stack.get_node_group_instances(node_group)
            for instance in instances:
                nova_id = instance['physical_id']
                if nova_id not in old_ids:
                    name = instance['name']
                    inst = {
                        "instance_id": nova_id,
                        "instance_name": name
                    }
                    if cluster.use_designate_feature():
                        inst.update(
                            {"dns_hostname":
                                name + '.' + cluster.domain_name[:-1]})
                    instance_id = conductor.instance_add(ctx, node_group, inst)
                    new_ids.append(instance_id)

        return new_ids

    def _rollback_cluster_creation(self, cluster, ex):
        """Shutdown all instances and update cluster status."""

        self.shutdown_cluster(cluster)

    def _rollback_cluster_scaling(self, cluster, rollback_count,
                                  target_count, ex):
        """Attempt to rollback cluster scaling.

        Our rollback policy for scaling is as follows:
        We shut down nodes created during scaling, but we don't try to
        to get back decommissioned nodes. I.e. during the rollback
        we only shut down nodes and not launch them. That approach should
        maximize the chance of rollback success.
        """

        for ng in rollback_count:
            if rollback_count[ng] > target_count[ng]:
                rollback_count[ng] = target_count[ng]

        self._launch_instances(cluster, rollback_count, ROLLBACK_STAGES,
                               update_stack=True)

    def shutdown_cluster(self, cluster, force=False):
        """Shutdown specified cluster and all related resources."""
        if force:
            heat_shutdown = heat.lazy_delete_stack
        else:
            heat_shutdown = heat.delete_stack

        try:
            heat_shutdown(cluster)
        except heat_exc.HTTPNotFound:
            LOG.warning('Did not find stack for cluster.')
        except ex.HeatStackException:
            raise

        self._clean_job_executions(cluster)
        self._remove_db_objects(cluster)

    @cpo.event_wrapper(
        True, step=_('Create Heat stack'), param=('cluster', 1))
    def _create_instances(self, cluster, target_count, update_stack=False,
                          disable_rollback=True, instances_to_delete=None):

        stack = ht.ClusterStack(cluster)

        self._update_instance_count(stack, cluster, target_count,
                                    instances_to_delete)
        stack.instantiate(update_existing=update_stack,
                          disable_rollback=disable_rollback)
        heat.wait_stack_completion(
            cluster, is_update=update_stack,
            last_updated_time=stack.last_updated_time)
        return self._populate_cluster(cluster, stack)

    def _launch_instances(self, cluster, target_count, stages,
                          update_stack=False, disable_rollback=True,
                          instances_to_delete=None):
        # create all instances
        cluster = c_u.change_cluster_status(cluster, stages[0])

        inst_ids = self._create_instances(
            cluster, target_count, update_stack, disable_rollback,
            instances_to_delete)

        # wait for all instances are up and networks ready
        cluster = c_u.change_cluster_status(cluster, stages[1])

        instances = c_u.get_instances(cluster, inst_ids)

        self._await_networks(cluster, instances)

        # prepare all instances
        cluster = c_u.change_cluster_status(cluster, stages[2])

        instances = c_u.get_instances(cluster, inst_ids)

        volumes.mount_to_instances(instances)

        self._configure_instances(cluster)

        return inst_ids

    def _update_instance_count(self, stack, cluster, target_count,
                               instances_to_delete=None):
        ctx = context.ctx()
        instances_name_to_delete = {}
        if instances_to_delete:
            for instance in instances_to_delete:
                node_group_id = instance['node_group']['id']
                if node_group_id not in instances_name_to_delete:
                    instances_name_to_delete[node_group_id] = []
                instances_name_to_delete[node_group_id].append(
                    instance['instance_name'])

        for node_group in cluster.node_groups:
            count = target_count[node_group.id]
            stack.add_node_group_extra(
                node_group.id, count, self._generate_user_data_script,
                instances_name_to_delete.get(node_group.id, None))

            for inst in node_group.instances:
                if (instances_to_delete and
                        node_group.id in instances_name_to_delete):
                    if (inst.instance_name in
                            instances_name_to_delete[node_group.id]):
                        conductor.instance_remove(ctx, inst)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


JOB_BINARY_UPDATE_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "minLength": 1,
            "maxLength": 50,
            "format": "valid_name"
        },
        "is_public": {
            "type": ["boolean", "null"],
        },
        "is_protected": {
            "type": ["boolean", "null"],
        }
    },
    "additionalProperties": False,
    "required": []
}



# Copyright 2015 Telles Nobrega
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Increase internal_ip and management_ip column size to work with IPv6

Revision ID: 025
Revises: 024
Create Date: 2015-07-17 09:58:22.128263

"""

# revision identifiers, used by Alembic.
revision = '025'
down_revision = '024'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('instances', 'internal_ip', type_=sa.String(45),
                    nullable=True)
    op.alter_column('instances', 'management_ip', type_=sa.String(45),
                    nullable=True)



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import random
import re
import string

import six

from sahara.plugins import base as plugins_base


@six.add_metaclass(abc.ABCMeta)
class DataSourceType(object):
    @plugins_base.required_with_default
    def construct_url(self, url, job_exec_id):
        """Resolve placeholders in the data source url

        Supported placeholders:
        * %RANDSTR(len)% - will be replaced with random string of lowercase
                           letters of length `len`
        * %JOB_EXEC_ID%  - will be replaced with the job execution ID

        :param url: String that represents an url with placeholders
        :param job_exec_id: Id of the job execution

        :returns: String that is an url without placeholders

        """
        def _randstr(match):
            random_len = int(match.group(1))
            return ''.join(random.choice(string.ascii_lowercase)
                           for _ in six.moves.range(random_len))

        url = url.replace("%JOB_EXEC_ID%", job_exec_id)
        url = re.sub(r"%RANDSTR\((\d+)\)%", _randstr, url)

        return url

    @plugins_base.required_with_default
    def prepare_cluster(self, data_source, cluster, **kwargs):
        """Makes a cluster ready to use this data source

        Different implementations for each data source, for HDFS
        will be configure the cluster, for Swift verify credentials,
        and so on

        :param data_source: The object handle to a data source
        :param cluster: The object handle to a cluster
        :returns: None
        """
        pass

    @plugins_base.required_with_default
    def get_runtime_url(self, url, cluster):
        """Get the runtime url of the data source for a cluster

        It will construct a runtime url if needed, if it's not needed
        it will use the native url as runtime url

        :param url: String that represents an already constructed url
        :param cluster: The object handle to a cluster
        :returns: String representing the runtime url
        """
        return url

    @plugins_base.required_with_default
    def get_urls(self, url, cluster, job_exec_id):
        """Get the native url and runtime url of a determined data source

        :param url: String that represents a url (constructed or not)
        :param cluster: The object handle to a cluster
        :param job_exec_id: Id of the job execution

        :returns: A tuple of the form (native_url, runtime_url), where the urls
        are Strings
        """
        native_url = self.construct_url(url, job_exec_id)
        runtime_url = self.get_runtime_url(native_url, cluster)
        return (native_url, runtime_url)

    @plugins_base.required_with_default
    def validate(self, data):
        """Method that validates the data passed through the API

        This method will be executed during the data source creation and
        update

        :raise: If data is invalid, InvalidDataException
        """
        pass

    @plugins_base.optional
    def _validate_url(self, url):
        """Auxiliary method used by the validate method"""
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2013 Mirantis Inc.
# Copyright (c) 2014 Adrien Vergé <adrien.verge@numergy.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cinderclient.v3 import client as cinder_client_v3
from keystoneauth1 import exceptions as keystone_exceptions
from oslo_config import cfg
from oslo_log import log as logging

from sahara import context
from sahara.service import sessions
from sahara.utils.openstack import base
from sahara.utils.openstack import keystone


LOG = logging.getLogger(__name__)


opts = [
    cfg.IntOpt('api_version', default=3,
               help='Version of the Cinder API to use.',
               deprecated_name='cinder_api_version'),
    cfg.BoolOpt('api_insecure',
                default=False,
                help='Allow to perform insecure SSL requests to cinder.'),
    cfg.StrOpt('ca_file',
               help='Location of ca certificates file to use for cinder '
                    'client requests.'),
    cfg.StrOpt("endpoint_type",
               default="internalURL",
               help="Endpoint type for cinder client requests")
]

cinder_group = cfg.OptGroup(name='cinder',
                            title='Cinder client options')

CONF = cfg.CONF
CONF.register_group(cinder_group)
CONF.register_opts(opts, group=cinder_group)


def validate_config():
    if CONF.cinder.api_version != 3:
        LOG.warning('Unsupported Cinder API version: {bad}. Please set a '
                    'correct value for cinder.api_version in your '
                    'sahara.conf file (currently supported versions are: '
                    '{supported}). Falling back to Cinder API version 3.'
                    .format(bad=CONF.cinder.api_version,
                            supported=[3]))
        CONF.set_override('api_version', 3, group='cinder')


def client():
    session = sessions.cache().get_session(sessions.SESSION_TYPE_CINDER)
    auth = keystone.auth()
    cinder = cinder_client_v3.Client(
        session=session, auth=auth,
        endpoint_type=CONF.cinder.endpoint_type,
        region_name=CONF.os_region_name)
    return cinder


def check_cinder_exists():
    service_type = 'volumev3'
    try:
        base.url_for(context.current().service_catalog, service_type,
                     endpoint_type=CONF.cinder.endpoint_type)
        return True
    except keystone_exceptions.EndpointNotFound:
        return False


def get_volume(volume_id):
    return base.execute_with_retries(client().volumes.get, volume_id)
